import base64
import time
import uuid
import zlib
from typing import Any, Dict, Tuple, Union

from tomodachi import logging
from tomodachi.envelope.proto_build.protobuf.sns_sqs_message_pb2 import SNSSQSMessage

PROTOCOL_VERSION = "tomodachi-protobuf-base--1.0.0"


class ProtobufBase(object):
    @classmethod
    def validate(cls, **kwargs: Any) -> None:
        if "proto_class" not in kwargs:
            raise Exception("No proto_class defined")
        if kwargs.get("proto_class", None).__class__.__name__ not in ("GeneratedProtocolMessageType", "MessageMeta"):
            from google.protobuf.message import Message  # isort: skip

            if not issubclass(kwargs.get("proto_class", None), Message):
                raise Exception("keyword argument 'proto_class' is not a protobuf message class")

    @classmethod
    async def build_message(cls, service: Any, topic: str, data: Any, **kwargs: Any) -> str:
        message_data = data.SerializeToString()

        data_encoding = "proto"
        if len(message_data) > 60000:
            message_data = zlib.compress(data.SerializeToString())
            data_encoding = "gzip_proto"

        message = SNSSQSMessage()
        message.service.name = str(getattr(service, "name", None) or "")
        message.service.uuid = str(getattr(service, "uuid", None) or "")
        message.metadata.message_uuid = "{}.{}".format(getattr(service, "uuid", ""), str(uuid.uuid4()))
        message.metadata.protocol_version = PROTOCOL_VERSION
        message.metadata.timestamp = time.time()
        message.metadata.topic = topic
        message.metadata.data_encoding = data_encoding
        message.data = message_data

        return base64.b64encode(message.SerializeToString()).decode("ascii")

    @classmethod
    async def parse_message(
        cls, payload: str, proto_class: Any = None, validator: Any = None, **kwargs: Any
    ) -> Union[Dict, Tuple]:
        message = SNSSQSMessage()
        message.ParseFromString(base64.b64decode(payload))

        message_uuid = message.metadata.message_uuid
        timestamp = message.metadata.timestamp

        raw_data = None
        obj = None

        if not proto_class:
            raw_data = message.data
        else:
            obj = proto_class()
            if message.metadata.data_encoding == "proto":
                obj.ParseFromString(message.data)
            elif message.metadata.data_encoding == "base64":  # deprecated
                obj.ParseFromString(base64.b64decode(message.data))
            elif message.metadata.data_encoding == "gzip_proto":
                obj.ParseFromString(zlib.decompress(message.data))
            elif message.metadata.data_encoding == "base64_gzip_proto":  # deprecated
                obj.ParseFromString(zlib.decompress(base64.b64decode(message.data)))
            elif message.metadata.data_encoding == "raw":
                raw_data = message.data

        if validator is not None:
            try:
                if hasattr(validator, "__func__"):
                    # for static functions
                    validator.__func__(obj)
                else:
                    # for non-static functions
                    validator(obj)
            except Exception as e:
                logging.getLogger("tomodachi.envelope").warning(e.__str__(), envelope="ProtobufBase")
                raise e

        return (
            {
                "service": {"name": message.service.name, "uuid": message.service.uuid},
                "metadata": {
                    "message_uuid": message.metadata.message_uuid,
                    "protocol_version": message.metadata.protocol_version,
                    "timestamp": message.metadata.timestamp,
                    "topic": message.metadata.topic,
                    "data_encoding": message.metadata.data_encoding,
                },
                "data": raw_data if raw_data is not None else obj,
            },
            message_uuid,
            timestamp,
        )


__all__ = [
    "PROTOCOL_VERSION",
    "ProtobufBase",
    "SNSSQSMessage",
]



from typing import List, Optional, Tuple, Union


def limit_exception_traceback(
    exc: BaseException, ignored_modules: Optional[Union[List[str], Tuple[str, ...]]] = None
) -> None:
    if not exc or not isinstance(exc, BaseException):
        return

    if not ignored_modules:
        return

    if getattr(exc, "__cause__", None):
        return

    frames = []
    tb_back = None
    tb = exc.__traceback__
    while tb:
        module_name = tb.tb_frame.f_globals.get("__name__", "") if tb.tb_frame and tb.tb_frame.f_globals else ""
        frames.append((tb, module_name))
        if module_name not in ignored_modules:
            tb_back = tb
        tb = tb.tb_next

    if not tb_back:
        return

    for tb, module_name in reversed(frames):
        if tb_back:
            if tb is not tb_back:
                continue
            tb_back = None

        if not tb.tb_next:
            continue

        if module_name in ignored_modules:
            exc.with_traceback(tb.tb_next)
            break



from typing import Any, Dict

__cached_defs: Dict[str, Any] = {}


def __getattr__(name: str) -> Any:
    if name in __cached_defs:
        return __cached_defs[name]

    import importlib  # noqa  # isort:skip

    if name in ("JsonBase", "ProtobufBase"):
        module = importlib.import_module("tomodachi.envelope")
    else:
        raise AttributeError("module 'tomodachi.protocol' has no attribute '{}'".format(name))

    __cached_defs[name] = getattr(module, name)
    return __cached_defs[name]


__all__ = ["JsonBase", "ProtobufBase"]



from __future__ import annotations

import asyncio
import functools
import inspect
import ipaddress
import os
import pathlib
import platform
import re
import time
import uuid
import warnings
from typing import Any, Callable, Dict, Iterable, List, Mapping, Optional, SupportsInt, Tuple, Union, cast

import yarl
from aiohttp import WSMsgType
from aiohttp import __version__ as aiohttp_version
from aiohttp import hdrs, web, web_protocol, web_server, web_urldispatcher
from aiohttp.helpers import BasicAuth
from aiohttp.http import HttpVersion
from aiohttp.streams import EofStream
from aiohttp.web_fileresponse import FileResponse
from multidict import CIMultiDict, CIMultiDictProxy

from tomodachi import get_contextvar, logging
from tomodachi._exception import limit_exception_traceback
from tomodachi.helpers.execution_context import (
    decrease_execution_context_value,
    increase_execution_context_value,
    set_execution_context,
)
from tomodachi.helpers.middleware import execute_middlewares
from tomodachi.invoker import Invoker
from tomodachi.options import Options


class HttpException(Exception):
    pass


def get_forwarded_remote_ip(request: web.BaseRequest) -> str:
    try:
        return cast(str, request._cache["forwarded_remote_ip"])
    except KeyError:
        return ""


class RequestHandler(web_protocol.RequestHandler):
    __slots__ = (
        *web_protocol.RequestHandler.__slots__,
        "_connection_start_time",
        "_keepalive",
    )

    _manager: Server

    @staticmethod
    def get_request_ip(request: web.Request, *a: Any, **kw: Any) -> str:
        warnings.warn(
            "Using the 'RequestHandler.get_request_ip()' function is deprecated. Use the 'tomodachi.get_forwarded_remote_ip()' function instead.",
            DeprecationWarning,
        )

        return get_forwarded_remote_ip(request)

    def _cache_remote_ip(self, request: web.BaseRequest) -> None:
        remote_ip: str = request.remote or ""

        if self._manager._real_ip_header and self._manager._real_ip_from:
            if any(
                [ipaddress.ip_address(remote_ip) in ipaddress.ip_network(cidr) for cidr in self._manager._real_ip_from]
            ):
                header_value = request.headers.get(self._manager._real_ip_header)
                if header_value:
                    remote_ip = header_value.split(",")[0].strip().split(" ")[0].strip()

        request._cache["forwarded_remote_ip"] = remote_ip
        request._cache["request_ip"] = remote_ip  # deprecated

    async def start(self) -> None:
        self._connection_start_time = time.time()
        await super().start()

    async def _handle_request(
        self,
        request: web.BaseRequest,
        start_time: float,
        *args: Any,
    ) -> Tuple[web.StreamResponse, bool]:
        self._cache_remote_ip(request)
        result: Tuple[web.StreamResponse, bool] = await super()._handle_request(request, start_time, *args)
        return result

    async def finish_response(self, request: web.BaseRequest, resp: web.StreamResponse, start_time: float) -> bool:
        result: bool = await super().finish_response(request, resp, start_time)
        return result

    def handle_error(
        self, request: Any, status: int = 500, exc: Any = None, message: Optional[str] = None
    ) -> web.Response:
        """Handle errors.

        Returns HTTP response with specific status code. Logs additional
        information. It always closes current connection."""
        if self.transport is None:
            # client has been disconnected during writing.
            if self._manager._access_log:
                request_ip = RequestHandler.get_request_ip(request, None)
                version_string = None
                if isinstance(request.version, HttpVersion):
                    version_string = "HTTP/{}.{}".format(request.version.major, request.version.minor)

                status_code = 499
                logging.getLogger("tomodachi.http.response").info(
                    "client disconnected during writing",
                    status_code=status_code,
                    remote_ip=request_ip or "",
                    auth_user=getattr(request._cache.get("auth") or {}, "login", None) or Ellipsis,
                    request_method=request.method,
                    request_path=request.path,
                    request_query_string=request.query_string or Ellipsis,
                    http_version=version_string,
                    request_content_length=request.content_length if request.content_length else Ellipsis,
                    user_agent=request.headers.get("User-Agent", ""),
                )

        headers: CIMultiDict = CIMultiDict({})
        headers[hdrs.CONTENT_TYPE] = "text/plain; charset=utf-8"

        msg = "" if status == 500 or not message else message

        headers[hdrs.CONTENT_LENGTH] = str(len(msg))
        headers[hdrs.SERVER] = self._manager._server_header

        if isinstance(request.version, HttpVersion) and (request.version.major, request.version.minor) in (
            (1, 0),
            (1, 1),
        ):
            headers[hdrs.CONNECTION] = "close"

        resp: web.Response = web.Response(status=status, text=msg, headers=headers)
        resp.force_close()

        # some data already got sent, connection is broken
        if request.writer.output_size > 0 or self.transport is None:
            self.force_close()
        elif self.transport is not None:
            request_ip = RequestHandler.get_request_ip(request, None)
            if not request_ip:
                peername = request.transport.get_extra_info("peername")
                if peername:
                    request_ip, _ = peername
            if self._manager._access_log:
                if not status or status >= 500:
                    logging.getLogger("tomodachi.http.response").warning(
                        "error in request handling",
                        status_code=status,
                        remote_ip=request_ip or "",
                        auth_user=getattr(request._cache.get("auth") or {}, "login", None) or Ellipsis,
                        response_content_length=len(msg),
                        request_content_length=(
                            request.content_length
                            if request.content_length
                            else (
                                len(request._read_bytes)
                                if request._read_bytes is not None and len(request._read_bytes)
                                else Ellipsis
                            )
                        ),
                        request_content_read_length=(
                            len(request._read_bytes)
                            if request._read_bytes is not None and len(request._read_bytes)
                            else ((request.content and getattr(request.content, "total_bytes", None)) or Ellipsis)
                        ),
                    )
                else:
                    logging.getLogger("tomodachi.http.response").info(
                        "bad http request",
                        status_code=status,
                        remote_ip=request_ip or "",
                        auth_user=getattr(request._cache.get("auth") or {}, "login", None) or Ellipsis,
                        response_content_length=len(msg),
                        request_content_length=(
                            request.content_length
                            if request.content_length
                            else (
                                len(request._read_bytes)
                                if request._read_bytes is not None and len(request._read_bytes)
                                else Ellipsis
                            )
                        ),
                        request_content_read_length=(
                            len(request._read_bytes)
                            if request._read_bytes is not None and len(request._read_bytes)
                            else ((request.content and getattr(request.content, "total_bytes", None)) or Ellipsis)
                        ),
                    )

        return resp


class Server(web_server.Server):
    __slots__ = (
        "_loop",
        "_connections",
        "_kwargs",
        "requests_count",
        "request_handler",
        "request_factory",
        "_server_header",
        "_access_log",
        "_real_ip_header",
        "_real_ip_from",
    )

    _loop: asyncio.AbstractEventLoop
    _kwargs: Dict[str, Any]

    _server_header: str
    _access_log: Union[bool, str]
    _real_ip_header: str
    _real_ip_from: List[str]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._server_header = cast(str, kwargs.pop("server_header", "") if kwargs else "")
        self._access_log = cast(Union[bool, str], kwargs.pop("access_log", False) if kwargs else False)
        self._real_ip_header = cast(str, kwargs.pop("real_ip_header", "") if kwargs else "")
        self._real_ip_from = cast(List[str], kwargs.pop("real_ip_from", []) if kwargs else [])

        kwargs["access_log"] = None

        super().__init__(*args, **kwargs)

    def __call__(self) -> RequestHandler:
        return RequestHandler(self, loop=self._loop, **self._kwargs)


class DynamicResource(web_urldispatcher.DynamicResource):
    def __init__(self, pattern: Any, *, name: Optional[str] = None) -> None:
        self._routes: List = []
        self._name = name
        self._pattern = pattern
        self._formatter = ""

        simplified = re.compile(r"^\^?(.+?)\$?$").match(pattern.pattern)
        self._simplified_pattern = simplified.group(1) if simplified else pattern.pattern


class Response(object):
    __slots__ = ("_body", "_status", "_reason", "_headers", "content_type", "charset", "missing_content_type")

    def __init__(
        self,
        *,
        body: Optional[Union[bytes, str]] = None,
        status: int = 200,
        reason: Optional[str] = None,
        headers: Optional[Union[Dict, CIMultiDict, CIMultiDictProxy]] = None,
        content_type: Optional[str] = None,
        charset: Optional[str] = None,
    ) -> None:
        if headers is None:
            headers = CIMultiDict()
        elif not isinstance(headers, (CIMultiDict, CIMultiDictProxy)):
            headers = CIMultiDict(headers)

        self._body = body
        self._status = status
        self._reason = reason
        self._headers = headers
        self.content_type = content_type if hdrs.CONTENT_TYPE not in headers else None
        self.charset = charset if hdrs.CONTENT_TYPE not in headers else None

        self.missing_content_type = hdrs.CONTENT_TYPE not in headers and not content_type and not charset

    def get_aiohttp_response(
        self, context: Dict, default_charset: Optional[str] = None, default_content_type: Optional[str] = None
    ) -> web.Response:
        if self.missing_content_type:
            self.charset = default_charset
            self.content_type = default_content_type

        charset = self.charset
        if hdrs.CONTENT_TYPE in self._headers and ";" in self._headers[hdrs.CONTENT_TYPE]:
            try:
                charset = (
                    str([v for v in self._headers[hdrs.CONTENT_TYPE].split(";") if "charset=" in v][0])
                    .replace("charset=", "")
                    .strip()
                )
            except IndexError:
                pass
        elif hdrs.CONTENT_TYPE in self._headers and ";" not in self._headers[hdrs.CONTENT_TYPE]:
            charset = None

        if self._body and not isinstance(self._body, bytes) and charset:
            body = self._body
            try:
                body_value = body.encode(charset.lower())
            except (ValueError, LookupError, UnicodeEncodeError) as e:
                logging.getLogger("exception").exception("uncaught exception: {}".format(str(e)))
                raise web.HTTPInternalServerError() from e
        elif self._body:
            body_value = self._body.encode() if not isinstance(self._body, bytes) else self._body
        else:
            body_value = b""

        response: web.Response = web.Response(
            body=body_value,
            status=self._status,
            reason=self._reason,
            headers=self._headers,
            content_type=self.content_type,
            charset=self.charset,
        )
        return response


class HttpTransport(Invoker):
    server_port_mapping: Dict[Any, str] = {}

    @classmethod
    async def request_handler(
        cls,
        obj: Any,
        context: Dict,
        func: Any,
        method: Union[str, List[str], Tuple[str, ...]],
        url: str,
        *,
        ignore_logging: Union[bool, List[int], Tuple[int, ...]] = False,
        pre_handler_func: Optional[Callable] = None,
    ) -> Any:
        pattern = r"^{}$".format(re.sub(r"\$$", "", re.sub(r"^\^?(.*)$", r"\1", url)))
        compiled_pattern = re.compile(pattern)

        http_options: Options.HTTP = cls.options(context).http
        default_content_type = http_options.content_type
        default_charset = http_options.charset
        if default_content_type is not None and ";" in default_content_type:
            # for backwards compability
            try:
                default_charset = (
                    str([v for v in default_content_type.split(";") if "charset=" in v][0])
                    .replace("charset=", "")
                    .strip()
                )
                default_content_type = str([v for v in default_content_type.split(";")][0]).strip()
            except IndexError:
                pass

        values = inspect.getfullargspec(func)
        original_kwargs = (
            {k: values.defaults[i] for i, k in enumerate(values.args[len(values.args) - len(values.defaults) :])}
            if values.defaults
            else {}
        )
        args_list = values.args[1 : len(values.args) - len(values.defaults or ())]
        args_set = (set(values.args[1:]) | set(values.kwonlyargs)) - set(["self"])

        middlewares = context.get("http_middleware", [])

        async def handler(request: web.Request) -> Union[web.Response, web.FileResponse]:
            logger = logging.getLogger("tomodachi.http.handler").bind(handler=func.__name__, type="tomodachi.http")

            kwargs = dict(original_kwargs)
            arg_matches: Dict[str, Any] = {}

            if "(" in pattern:
                result = compiled_pattern.match(request.path)
                if result:
                    for k, v in result.groupdict().items():
                        if k in args_set:
                            kwargs[k] = v
                            if k in values.args:
                                arg_matches[k] = v

            if "request" in args_set:
                kwargs["request"] = request
                if "request" in values.args:
                    arg_matches["request"] = request

            if not context.get("_http_accept_new_requests"):
                raise web.HTTPServiceUnavailable()

            if pre_handler_func:
                await pre_handler_func(obj, request)
                logger = logging.getLogger("tomodachi.http.handler")

            @functools.wraps(func)
            async def routine_func(
                *a: Any, **kw: Any
            ) -> Union[str, bytes, Dict, List, Tuple, web.Response, web.FileResponse, Response]:
                logging.bind_logger(logger)
                get_contextvar("service.logger").set("tomodachi.http.handler")

                kw_values = {k: v for k, v in {**kwargs, **kw}.items() if values.varkw or k in args_set}
                args_values = [
                    kw_values.pop(key) if key in kw_values else a[i + 1]
                    for i, key in enumerate(values.args[1 : len(a) + 1])
                ]
                if values.varargs and not values.defaults and len(a) > len(args_values) + 1:
                    args_values += a[len(args_values) + 1 :]

                routine = func(*(obj, *args_values), **kw_values)
                return_value: Union[str, bytes, Dict, List, Tuple, web.Response, web.FileResponse, Response] = (
                    (await routine) if inspect.isawaitable(routine) else routine
                )
                return return_value

            return_value: Union[str, bytes, Dict, List, Tuple, web.Response, web.FileResponse, Response]
            if middlewares:
                logging.bind_logger(
                    logging.getLogger("tomodachi.http.middleware").bind(
                        middleware=Ellipsis, handler=func.__name__, type="tomodachi.http"
                    )
                )
                return_value = await asyncio.create_task(
                    execute_middlewares(func, routine_func, middlewares, *(obj, request), request=request)
                )
            else:
                logging.bind_logger(logger)
                get_contextvar("service.logger").set("tomodachi.http.handler")

                a = [arg_matches[k] if k in arg_matches else (request,)[i] for i, k in enumerate(args_list)]
                args_values = [kwargs.pop(key) if key in kwargs else a[i] for i, key in enumerate(args_list)]

                if values.varargs and not values.defaults and len(a) > len(args_values) + 1:
                    args_values += a[len(args_values) + 1 :]

                routine = func(obj, *args_values, **kwargs)
                return_value = (await routine) if inspect.isawaitable(routine) else routine

            response = resolve_response_sync(
                return_value,
                request=request,
                context=context,
                default_content_type=default_content_type,
                default_charset=default_charset,
            )
            return response

        context["_http_routes"] = context.get("_http_routes", [])
        route_context = {"ignore_logging": ignore_logging}
        if isinstance(method, list) or isinstance(method, tuple):
            for m in method:
                context["_http_routes"].append((m.upper(), pattern, handler, route_context))
        elif isinstance(method, str):
            context["_http_routes"].append((method.upper(), pattern, handler, route_context))
        else:
            raise Exception("Invalid method '{}' for route".format(str(method)))

        start_func = cls.start_server(obj, context)
        return (await start_func) if start_func else None

    @classmethod
    async def static_request_handler(
        cls,
        obj: Any,
        context: Dict,
        func: Any,
        path: str,
        base_url: str,
        *,
        ignore_logging: Union[bool, List[int], Tuple[int, ...]] = False,
    ) -> Any:
        if "?P<filename>" not in base_url:
            pattern = r"^{}(?P<filename>.+?)$".format(re.sub(r"\$$", "", re.sub(r"^\^?(.*)$", r"\1", base_url)))
        else:
            pattern = r"^{}$".format(re.sub(r"\$$", "", re.sub(r"^\^?(.*)$", r"\1", base_url)))
        compiled_pattern = re.compile(pattern)

        if path in ("", "/"):
            # Hopefully noone wants to do this intentionally, and if anyone accidentally does we'll catch it here
            raise Exception("Invalid path '{}' for static route".format(path))

        if not path.startswith("/"):
            path = "{}/{}".format(os.path.dirname(context.get("context", {}).get("_service_file_path", "")), path)

        if not path.endswith("/"):
            path = "{}/".format(path)

        if os.path.realpath(path) == "/":
            raise Exception("Invalid path '{}' for static route resolves to '/'".format(path))

        async def handler(request: web.Request) -> Union[web.Response, web.FileResponse]:
            normalized_request_path = yarl.URL._normalize_path(request.path)
            if not normalized_request_path.startswith("/"):
                raise web.HTTPNotFound()

            result = compiled_pattern.match(normalized_request_path)
            filename = result.groupdict()["filename"] if result else ""

            basepath = os.path.realpath(path)
            realpath = os.path.realpath("{}/{}".format(basepath, filename))

            try:
                if any(
                    [
                        not realpath,
                        not basepath,
                        realpath == basepath,
                        os.path.commonprefix((realpath, basepath)) != basepath,
                        not os.path.exists(realpath),
                        not os.path.isdir(basepath),
                        basepath == "/",
                        os.path.isdir(realpath),
                    ]
                ):
                    raise web.HTTPNotFound()

                # deepcode ignore PT: Input data to pathlib.Path is sanitized
                pathlib.Path(realpath).open("r")

                response: Union[web.Response, web.FileResponse] = FileResponse(path=realpath, chunk_size=256 * 1024)
                return response
            except PermissionError:
                raise web.HTTPForbidden()

        route_context = {"ignore_logging": ignore_logging}
        context["_http_routes"] = context.get("_http_routes", [])
        context["_http_routes"].append(("GET", pattern, handler, route_context))

        start_func = cls.start_server(obj, context)
        return (await start_func) if start_func else None

    @classmethod
    async def error_handler(cls, obj: Any, context: Dict, func: Any, status_code: int) -> Any:
        http_options: Options.HTTP = cls.options(context).http
        default_content_type = http_options.content_type
        default_charset = http_options.charset
        if default_content_type is not None and ";" in default_content_type:
            # for backwards compability
            try:
                default_charset = (
                    str([v for v in default_content_type.split(";") if "charset=" in v][0])
                    .replace("charset=", "")
                    .strip()
                )
                default_content_type = str([v for v in default_content_type.split(";")][0]).strip()
            except IndexError:
                pass

        values = inspect.getfullargspec(func)
        original_kwargs = (
            {k: values.defaults[i] for i, k in enumerate(values.args[len(values.args) - len(values.defaults) :])}
            if values.defaults
            else {}
        )
        args_list = values.args[1 : len(values.args) - len(values.defaults or ())]
        args_set = (set(values.args[1:]) | set(values.kwonlyargs)) - set(["self"])

        middlewares = context.get("http_middleware", [])

        async def handler(request: web.Request) -> Union[web.Response, web.FileResponse]:
            logger = logging.getLogger("tomodachi.http.handler").bind(
                handler=func.__name__, type="tomodachi.http_error", status_code=status_code
            )

            kwargs = dict(original_kwargs)
            arg_matches: Dict[str, Any] = {}

            if "request" in args_set:
                kwargs["request"] = request
                if "request" in values.args:
                    arg_matches["request"] = request

            if "status_code" in args_set:
                kwargs["status_code"] = status_code
                if "status_code" in values.args:
                    arg_matches["status_code"] = status_code

            request._cache["error_status_code"] = status_code

            @functools.wraps(func)
            async def routine_func(
                *a: Any, **kw: Any
            ) -> Union[str, bytes, Dict, List, Tuple, web.Response, web.FileResponse, Response]:
                logging.bind_logger(logger)
                get_contextvar("service.logger").set("tomodachi.http.handler")

                kw_values = {k: v for k, v in {**kwargs, **kw}.items() if values.varkw or k in args_set}
                args_values = [
                    kw_values.pop(key) if key in kw_values else a[i + 1]
                    for i, key in enumerate(values.args[1 : len(a) + 1])
                ]
                if values.varargs and not values.defaults and len(a) > len(args_values) + 1:
                    args_values += a[len(args_values) + 1 :]

                routine = func(*(obj, *args_values), **kw_values)
                return_value: Union[str, bytes, Dict, List, Tuple, web.Response, Response] = (
                    (await routine) if inspect.isawaitable(routine) else routine
                )
                return return_value

            return_value: Union[str, bytes, Dict, List, Tuple, web.Response, web.FileResponse, Response]
            if middlewares:
                logging.bind_logger(
                    logging.getLogger("tomodachi.http.middleware").bind(
                        middleware=Ellipsis, handler=func.__name__, type="tomodachi.http_error", status_code=status_code
                    )
                )
                return_value = await asyncio.create_task(
                    execute_middlewares(
                        func, routine_func, middlewares, *(obj, request), request=request, status_code=status_code
                    )
                )
            else:
                logging.bind_logger(logger)
                get_contextvar("service.logger").set("tomodachi.http.handler")

                a = [arg_matches[k] if k in arg_matches else (request,)[i] for i, k in enumerate(args_list)]
                args_values = [kwargs.pop(key) if key in kwargs else a[i] for i, key in enumerate(args_list)]

                if values.varargs and not values.defaults and len(a) > len(args_values) + 1:
                    args_values += a[len(args_values) + 1 :]

                routine = func(obj, *args_values, **kwargs)
                return_value = (await routine) if inspect.isawaitable(routine) else routine

            response = resolve_response_sync(
                return_value,
                request=request,
                context=context,
                status_code=status_code,
                default_content_type=default_content_type,
                default_charset=default_charset,
            )
            return response

        context["_http_error_handler"] = context.get("_http_error_handler", {})
        context["_http_error_handler"][int(status_code)] = handler

        start_func = cls.start_server(obj, context)
        return (await start_func) if start_func else None

    @classmethod
    async def websocket_handler(cls, obj: Any, context: Dict, func: Any, url: str) -> Any:
        response_logger = logging.getLogger("tomodachi.http.websocket")

        pattern = r"^{}$".format(re.sub(r"\$$", "", re.sub(r"^\^?(.*)$", r"\1", url)))
        compiled_pattern = re.compile(pattern)

        access_log = cls.options(context).http.access_log

        async def _pre_handler_func(_: Any, request: web.Request) -> None:
            request._cache["is_websocket"] = True
            request._cache["websocket_uuid"] = str(uuid.uuid4())

            logging.getLogger("tomodachi.http.handler").bind(handler=func.__name__, type="tomodachi.websocket")

        values = inspect.getfullargspec(func)
        original_kwargs = (
            {k: values.defaults[i] for i, k in enumerate(values.args[len(values.args) - len(values.defaults) :])}
            if values.defaults
            else {}
        )
        args_list = values.args[1 : len(values.args) - len(values.defaults or ())]
        args_set = (set(values.args[1:]) | set(values.kwonlyargs)) - set(["self"])

        @functools.wraps(func)
        async def _func(obj: Any, request: web.Request, *_a: Any, **kw: Any) -> None:
            websocket = web.WebSocketResponse()

            request_ip = RequestHandler.get_request_ip(request, context)
            try:
                await websocket.prepare(request)
            except Exception:
                try:
                    await websocket.close()
                except Exception:
                    pass

                if access_log:
                    response_logger.info(
                        websocket_state="cancelled",
                        remote_ip=request_ip,
                        auth_user=getattr(request._cache.get("auth") or {}, "login", None) or Ellipsis,
                        request_path=request.path,
                        request_query_string=request.query_string or Ellipsis,
                        websocket_id=request._cache.get("websocket_uuid", ""),
                        user_agent=request.headers.get("User-Agent", ""),
                    )

                return

            context["_http_open_websockets"] = context.get("_http_open_websockets", [])
            context["_http_open_websockets"].append(websocket)

            if access_log:
                response_logger.info(
                    websocket_state="open",
                    remote_ip=request_ip,
                    auth_user=getattr(request._cache.get("auth") or {}, "login", None) or Ellipsis,
                    request_path=request.path,
                    request_query_string=request.query_string or Ellipsis,
                    websocket_id=request._cache.get("websocket_uuid", ""),
                    user_agent=request.headers.get("User-Agent", ""),
                )

            kwargs = dict(original_kwargs)
            arg_matches: Dict[str, Any] = {}

            if "(" in pattern:
                result = compiled_pattern.match(request.path)
                if result:
                    for k, v in result.groupdict().items():
                        if k in args_set:
                            kwargs[k] = v
                            if k in values.args:
                                arg_matches[k] = v

            if "request" in args_set:
                kwargs["request"] = request
                if "request" in values.args:
                    arg_matches["request"] = request

            if "websocket" in args_set:
                kwargs["websocket"] = websocket
                if "websocket" in values.args:
                    arg_matches["websocket"] = websocket

            try:
                kw_values = {k: v for k, v in {**kwargs, **kw}.items() if values.varkw or k in args_set}
                a = [arg_matches[k] if k in arg_matches else (websocket, request)[i] for i, k in enumerate(args_list)]
                args_values = [kw_values.pop(key) if key in kw_values else a[i] for i, key in enumerate(args_list)]

                if values.varargs and not values.defaults and len(a) > len(args_values) + 1:
                    args_values += a[len(args_values) + 1 :]

                routine = func(obj, *args_values, **kw_values)
                callback_functions: Optional[Union[Tuple[Callable, Callable], Tuple[Callable], Callable]] = (
                    (await routine) if inspect.isawaitable(routine) else routine
                )
            except Exception as e:
                limit_exception_traceback(e, ("tomodachi.transport.http", "tomodachi.helpers.middleware"))
                logging.getLogger("exception").exception("uncaught exception: {}".format(str(e)))
                try:
                    await websocket.close()
                except Exception:
                    pass

                try:
                    context["_http_open_websockets"].remove(websocket)
                except Exception:
                    pass

                if access_log:
                    response_logger.info(
                        websocket_state="error",
                        remote_ip=request_ip,
                        auth_user=getattr(request._cache.get("auth") or {}, "login", None) or Ellipsis,
                        request_path=request.path,
                        request_query_string=request.query_string or Ellipsis,
                        websocket_id=request._cache.get("websocket_uuid", ""),
                        user_agent=request.headers.get("User-Agent", ""),
                    )

                return

            _receive_func = None
            _close_func = None

            if callback_functions and isinstance(callback_functions, tuple):
                if len(callback_functions) == 2:
                    _receive_func, _close_func = callback_functions
                elif len(callback_functions) == 1:
                    (_receive_func,) = callback_functions
            elif callback_functions:
                _receive_func = callback_functions

            try:
                async for message in websocket:
                    if message.type == WSMsgType.TEXT:
                        if _receive_func:
                            try:
                                await _receive_func(message.data)
                            except Exception as e:
                                limit_exception_traceback(e, ("tomodachi.transport.http",))
                                logging.getLogger("exception").exception("uncaught exception: {}".format(str(e)))
                    elif message.type == WSMsgType.ERROR:
                        if not context.get("log_level") or context.get("log_level") in ["DEBUG"]:
                            ws_exception = websocket.exception()
                            if isinstance(ws_exception, (EofStream, RuntimeError)):
                                pass
                            elif isinstance(ws_exception, Exception):
                                limit_exception_traceback(ws_exception, ("tomodachi.transport.http",))
                                logging.getLogger("exception").exception(
                                    "uncaught exception: {}".format(str(ws_exception)), exception=ws_exception
                                )
                            else:
                                response_logger.warning("websocket exception", websocket_exception=ws_exception)
                    elif message.type == WSMsgType.CLOSED:
                        break  # noqa
            except Exception:
                pass
            finally:
                if _close_func:
                    try:
                        await _close_func()
                    except Exception as e:
                        limit_exception_traceback(e, ("tomodachi.transport.http",))
                        logging.getLogger("exception").exception("uncaught exception: {}".format(str(e)))
                try:
                    await websocket.close()
                except Exception:
                    pass

                try:
                    context["_http_open_websockets"].remove(websocket)
                except Exception:
                    pass

        return await cls.request_handler(obj, context, _func, "GET", url, pre_handler_func=_pre_handler_func)

    @staticmethod
    async def start_server(obj: Any, context: Dict) -> Optional[Callable]:
        if context.get("_http_server_started"):
            return None
        context["_http_server_started"] = True

        logger = logging.getLogger("tomodachi.http")
        logging.bind_logger(logger)

        http_options: Options.HTTP = HttpTransport.options(context).http

        server_header = http_options.server_header or ""
        access_log = http_options.access_log or False
        real_ip_header = http_options.real_ip_header or ""
        real_ip_from = (
            [http_options.real_ip_from]
            if http_options.real_ip_from and isinstance(http_options.real_ip_from, str)
            else http_options.real_ip_from or []
        )

        logger_handler = None
        if isinstance(access_log, str):
            from logging.handlers import WatchedFileHandler  # noqa  # isort:skip

            try:
                wfh = WatchedFileHandler(filename=access_log)
            except FileNotFoundError as e:
                logger.warning('Unable to use file for access log - invalid path ("{}")'.format(access_log))
                raise HttpException(str(e)) from e
            except PermissionError as e:
                logger.warning('Unable to use file for access log - invalid permissions ("{}")'.format(access_log))
                raise HttpException(str(e)) from e
            wfh.setLevel(logging.DEBUG)
            wfh.setFormatter(logging.JSONFormatter)
            logger_handler = wfh
            logger.info("logging requests to file", file_path=access_log)
            logging.getLogger("tomodachi.http.response").addHandler(logger_handler)

        async def _start_server() -> None:
            logger = logging.getLogger("tomodachi.http")
            logging.bind_logger(logger)
            response_logger = logging.getLogger("tomodachi.http.response")

            loop = asyncio.get_event_loop()

            logging.getLogger("aiohttp.access").setLevel(logging.WARNING)

            async def request_handler_func(
                request: web.Request, handler: Callable, request_start_time: int = 0
            ) -> Union[web.Response, web.FileResponse]:
                response: Optional[Union[web.Response, web.FileResponse]] = None
                request_ip = RequestHandler.get_request_ip(request, context)

                # try to read body if it exists and can be read
                premature_eof = False
                if request.body_exists and request.can_read_body and (request.content_length or request.content):
                    try:
                        if (
                            request._read_bytes is None
                            and request.content
                            and request.content.is_eof()
                            and getattr(request.content, "_size", 0) > 0
                            and request.content.exception()
                        ):
                            request._read_bytes = request.content._read_nowait(-1)
                        else:
                            await request.read()
                    except web.HTTPException as exc:
                        # internal aiohttp exception raised (for example if entity too large)
                        response = exc
                    except Exception:
                        # failed to read body (for example if connection is closed before the entire body was sent)
                        premature_eof = True

                if request.headers.get("Authorization"):
                    try:
                        request._cache["auth"] = BasicAuth.decode(request.headers.get("Authorization", ""))
                    except ValueError:
                        pass

                if not request_ip or premature_eof:
                    # ignore request for broken transport before request handling and before entire body was sent
                    response = web.Response(status=499, headers={hdrs.SERVER: server_header})
                    response._eof_sent = True
                    response.force_close()

                    if access_log:
                        status_code = response.status if response is not None else 500
                        ignore_logging = getattr(handler, "ignore_logging", False)
                        if ignore_logging is True:
                            pass
                        elif isinstance(ignore_logging, (list, tuple)) and status_code in ignore_logging:
                            pass

                        request_version = (
                            (request.version.major, request.version.minor)
                            if isinstance(request.version, HttpVersion)
                            else (1, 0)
                        )
                        version_string = None
                        if isinstance(request.version, HttpVersion):
                            version_string = "HTTP/{}.{}".format(request_version[0], request_version[1])

                        msg = (
                            "ignored http request - connection closed before body was sent"
                            if premature_eof
                            else "ignored http request"
                        )
                        response_logger.info(
                            msg,
                            status_code=499,
                            remote_ip=request_ip or "",
                            auth_user=getattr(request._cache.get("auth") or {}, "login", None) or Ellipsis,
                            request_method=request.method,
                            request_path=request.path,
                            request_query_string=request.query_string or Ellipsis,
                            http_version=version_string,
                            response_content_length=(
                                response.content_length
                                if response is not None and response.content_length is not None
                                else Ellipsis
                            ),
                            request_content_length=(
                                request.content_length
                                if request.content_length
                                else (
                                    len(request._read_bytes)
                                    if not premature_eof
                                    and request._read_bytes is not None
                                    and len(request._read_bytes)
                                    else Ellipsis
                                )
                            ),
                            request_content_read_length=(
                                len(request._read_bytes)
                                if request._read_bytes is not None and len(request._read_bytes)
                                else ((request.content and getattr(request.content, "total_bytes", None)) or Ellipsis)
                            ),
                            user_agent=request.headers.get("User-Agent", ""),
                        )

                    return response

                handler_start_time: int = 0
                handler_stop_time: int = 0

                caught_exceptions: List[Exception] = []

                try:
                    if not response:
                        handler_start_time = time.perf_counter_ns() if access_log else 0
                        response = await handler(request)
                        handler_stop_time = time.perf_counter_ns() if access_log else 0
                except web.HTTPException as e:
                    handler_stop_time = time.perf_counter_ns() if access_log else 0
                    error_handler = context.get("_http_error_handler", {}).get(e.status, None)
                    if error_handler:
                        try:
                            response = await error_handler(request)
                        except Exception as error_handler_exception:
                            limit_exception_traceback(
                                error_handler_exception, ("tomodachi.transport.http", "tomodachi.helpers.middleware")
                            )
                            logging.getLogger("exception").exception(
                                "uncaught exception: {}".format(str(error_handler_exception))
                            )
                            caught_exceptions.append(error_handler_exception)
                            error_handler = context.get("_http_error_handler", {}).get(500, None)
                            if error_handler:
                                try:
                                    response = await error_handler(request)
                                except Exception as fallback_error_handler_exception:
                                    limit_exception_traceback(
                                        fallback_error_handler_exception,
                                        ("tomodachi.transport.http", "tomodachi.helpers.middleware"),
                                    )
                                    logging.getLogger("exception").exception(
                                        "uncaught exception: {}".format(str(fallback_error_handler_exception))
                                    )
                                    caught_exceptions.append(fallback_error_handler_exception)
                                    response = web.HTTPInternalServerError()
                                    response.body = b""
                            else:
                                response = web.HTTPInternalServerError()
                                response.body = b""
                    else:
                        response = e
                        response.body = str(e).encode("utf-8")
                except Exception as e:
                    handler_stop_time = time.perf_counter_ns() if access_log else 0
                    limit_exception_traceback(e, ("tomodachi.transport.http", "tomodachi.helpers.middleware"))
                    logging.getLogger("exception").exception("uncaught exception: {}".format(str(e)))
                    caught_exceptions.append(e)
                    error_handler = context.get("_http_error_handler", {}).get(500, None)
                    if error_handler:
                        try:
                            response = await error_handler(request)
                        except Exception as fallback_error_handler_exception:
                            limit_exception_traceback(
                                fallback_error_handler_exception,
                                ("tomodachi.transport.http", "tomodachi.helpers.middleware"),
                            )
                            logging.getLogger("exception").exception(
                                "uncaught exception: {}".format(str(fallback_error_handler_exception))
                            )
                            caught_exceptions.append(fallback_error_handler_exception)
                            response = web.HTTPInternalServerError()
                            response.body = b""

                    else:
                        response = web.HTTPInternalServerError()
                        response.body = b""
                finally:
                    replaced_status_code = None
                    replaced_response_content_length = None
                    if not request.transport:
                        replaced_status_code = response.status if response is not None else 500
                        replaced_response_content_length = (
                            response.content_length
                            if response is not None and response.content_length is not None
                            else None
                        )

                        response = web.Response(status=499, headers={})
                        response._eof_sent = True
                        setattr(response, "_replaced_status_code", replaced_status_code)
                        if replaced_response_content_length is not None:
                            setattr(response, "_replaced_response_content_length", replaced_response_content_length)

                    request_version = (
                        (request.version.major, request.version.minor)
                        if isinstance(request.version, HttpVersion)
                        else (1, 0)
                    )

                    if access_log:
                        total_request_time = ((time.perf_counter_ns() - request_start_time) / 1000000.0) / 1000.0
                        handler_elapsed_time = ((handler_stop_time - handler_start_time) / 1000000.0) / 1000.0

                        version_string = None
                        if isinstance(request.version, HttpVersion):
                            version_string = "HTTP/{}.{}".format(request_version[0], request_version[1])

                        if not request._cache.get("is_websocket"):
                            status_code = response.status if response is not None else 500
                            ignore_logging = getattr(handler, "ignore_logging", False)
                            if ignore_logging is True:
                                pass
                            elif isinstance(ignore_logging, (list, tuple)) and status_code in ignore_logging:
                                pass
                            else:
                                response_logger.info(
                                    status_code=status_code,
                                    replaced_status_code=replaced_status_code or Ellipsis,
                                    remote_ip=request_ip,
                                    auth_user=getattr(request._cache.get("auth") or {}, "login", None) or Ellipsis,
                                    request_method=request.method,
                                    request_path=request.path,
                                    request_query_string=request.query_string or Ellipsis,
                                    http_version=version_string,
                                    response_content_length=(
                                        response.content_length
                                        if response is not None and response.content_length is not None
                                        else Ellipsis
                                    ),
                                    replaced_response_content_length=replaced_response_content_length or Ellipsis,
                                    request_content_length=(
                                        request.content_length
                                        if request.content_length
                                        else (
                                            len(request._read_bytes)
                                            if request._read_bytes is not None and len(request._read_bytes)
                                            else Ellipsis
                                        )
                                    ),
                                    request_content_read_length=(
                                        len(request._read_bytes)
                                        if request._read_bytes is not None and len(request._read_bytes)
                                        else (
                                            (request.content and getattr(request.content, "total_bytes", None))
                                            or Ellipsis
                                        )
                                    ),
                                    user_agent=request.headers.get("User-Agent", ""),
                                    handler_elapsed_time=(
                                        "{0:.5f}s".format(round(handler_elapsed_time, 5))
                                        if handler_start_time and handler_stop_time
                                        else Ellipsis
                                    ),
                                    request_time="{0:.5f}s".format(round(total_request_time, 5)),
                                )
                        else:
                            response_logger.info(
                                websocket_state="closed",
                                remote_ip=request_ip,
                                auth_user=getattr(request._cache.get("auth") or {}, "login", None) or Ellipsis,
                                request_path=request.path,
                                request_query_string=request.query_string or Ellipsis,
                                websocket_id=request._cache.get("websocket_uuid", ""),
                                user_agent=request.headers.get("User-Agent", ""),
                                total_request_time="{0:.5f}s".format(round(total_request_time, 5)),
                            )

                    if response is not None:
                        response.headers[hdrs.SERVER] = server_header

                        if request_version in ((1, 0), (1, 1)) and not request._cache.get("is_websocket"):
                            use_keepalive = False
                            if context["_http_tcp_keepalive"] and request.keep_alive and request.protocol:
                                use_keepalive = True
                                if any(
                                    [
                                        # keep-alive timeout not set or is non-positive
                                        (
                                            not context["_http_keepalive_timeout"]
                                            or context["_http_keepalive_timeout"] <= 0
                                        ),
                                        # keep-alive request count has passed configured max for this connection
                                        (
                                            context["_http_max_keepalive_requests"]
                                            and request.protocol._request_count
                                            >= context["_http_max_keepalive_requests"]
                                        ),
                                        # keep-alive time has passed configured max for this connection
                                        (
                                            context["_http_max_keepalive_time"]
                                            and time.time()
                                            > getattr(request.protocol, "_connection_start_time", 0)
                                            + context["_http_max_keepalive_time"]
                                        ),
                                    ]
                                ):
                                    use_keepalive = False

                            if use_keepalive:
                                response.headers[hdrs.CONNECTION] = "keep-alive"
                                response.headers[hdrs.KEEP_ALIVE] = "timeout={}{}".format(
                                    request.protocol._keepalive_timeout,
                                    (
                                        ", max={}".format(context["_http_max_keepalive_requests"])
                                        if context["_http_max_keepalive_requests"]
                                        else ""
                                    ),
                                )
                            else:
                                response.headers[hdrs.CONNECTION] = "close"
                                response.force_close()

                        if not context["_http_tcp_keepalive"] and not request._cache.get("is_websocket"):
                            response.force_close()

                    if response is None:
                        try:
                            raise Exception("invalid response value")
                        except Exception as e:
                            limit_exception_traceback(e, ("tomodachi.transport.http",))
                            logging.getLogger("exception").exception("uncaught exception: {}".format(str(e)))

                        response = web.HTTPInternalServerError()
                        response.body = b""
                        response.headers[hdrs.CONNECTION] = "close"
                        response.force_close()

                    if caught_exceptions:
                        setattr(response, "_caught_exceptions", caught_exceptions)

                    if isinstance(response, (web.HTTPException, web.HTTPInternalServerError)):
                        raise response

                    return response

            @web.middleware
            async def middleware(request: web.Request, handler: Callable) -> Union[web.Response, web.FileResponse]:
                request_start_time = time.perf_counter_ns() if access_log else 0

                increase_execution_context_value("http_current_tasks")
                increase_execution_context_value("http_total_tasks")

                task = asyncio.ensure_future(
                    request_handler_func(request, handler, request_start_time=request_start_time)
                )
                context["_http_active_requests"] = context.get("_http_active_requests", set())
                context["_http_active_requests"].add(task)
                try:
                    await asyncio.shield(task)
                except asyncio.CancelledError:
                    try:
                        await task
                        decrease_execution_context_value("http_current_tasks")
                        try:
                            context["_http_active_requests"].remove(task)
                        except KeyError:
                            pass
                        return task.result()
                    except Exception:
                        decrease_execution_context_value("http_current_tasks")
                        try:
                            context["_http_active_requests"].remove(task)
                        except KeyError:
                            pass
                        raise
                except web.HTTPException:
                    decrease_execution_context_value("http_current_tasks")
                    try:
                        context["_http_active_requests"].remove(task)
                    except KeyError:
                        pass
                    raise
                except Exception as e:
                    decrease_execution_context_value("http_current_tasks")
                    try:
                        context["_http_active_requests"].remove(task)
                    except KeyError:
                        pass
                    limit_exception_traceback(e, ("tomodachi.transport.http",))
                    logging.getLogger("exception").exception("uncaught exception: {}".format(str(e)))
                    raise
                except BaseException:
                    decrease_execution_context_value("http_current_tasks")
                    try:
                        context["_http_active_requests"].remove(task)
                    except KeyError:
                        pass
                    raise
                decrease_execution_context_value("http_current_tasks")
                try:
                    context["_http_active_requests"].remove(task)
                except KeyError:
                    pass

                return task.result()

            client_max_size_option = http_options.client_max_size
            client_max_size_option_str = str(client_max_size_option).upper()
            client_max_size = (1024**2) * 100
            try:
                if (
                    client_max_size_option
                    and isinstance(client_max_size_option, str)
                    and (client_max_size_option_str.endswith("G") or client_max_size_option_str.endswith("GB"))
                ):
                    client_max_size = int(
                        re.sub(cast(str, r"^([0-9]+)GB?$"), cast(str, r"\1"), client_max_size_option_str)
                    ) * (1024**3)
                elif (
                    client_max_size_option
                    and isinstance(client_max_size_option, str)
                    and (client_max_size_option_str.endswith("M") or client_max_size_option_str.endswith("MB"))
                ):
                    client_max_size = int(re.sub(r"^([0-9]+)MB?$", r"\1", client_max_size_option_str)) * (1024**2)
                elif (
                    client_max_size_option
                    and isinstance(client_max_size_option, str)
                    and (client_max_size_option_str.endswith("K") or client_max_size_option_str.endswith("KB"))
                ):
                    client_max_size = int(re.sub(r"^([0-9]+)KB?$", r"\1", client_max_size_option_str)) * 1024
                elif (
                    client_max_size_option
                    and isinstance(client_max_size_option, str)
                    and (client_max_size_option_str.endswith("B"))
                ):
                    client_max_size = int(re.sub(r"^([0-9]+)B?$", r"\1", client_max_size_option_str))
                elif client_max_size_option:
                    client_max_size = int(client_max_size_option)
            except Exception:
                raise ValueError(
                    "Bad value for http option client_max_size: {}".format(str(client_max_size_option))
                ) from None
            if client_max_size >= 0 and client_max_size < 1024:
                raise ValueError(
                    "Too low value for http option client_max_size: {} ({})".format(
                        str(client_max_size_option), client_max_size_option
                    )
                )
            if client_max_size > 1024**3:
                raise ValueError(
                    "Too high value for http option client_max_size: {} ({})".format(
                        str(client_max_size_option), client_max_size_option
                    )
                )

            middlewares = context.get("_aiohttp_pre_middleware", []) + [middleware]
            app: web.Application = web.Application(middlewares=middlewares, client_max_size=client_max_size)
            app._set_loop(None)
            for method, pattern, handler, route_context in context.get("_http_routes", []):
                try:
                    compiled_pattern = re.compile(pattern)
                except re.error as exc:
                    raise ValueError("Bad http route pattern '{}': {}".format(pattern, exc)) from None
                ignore_logging = route_context.get("ignore_logging", False)
                setattr(handler, "ignore_logging", ignore_logging)
                resource = DynamicResource(compiled_pattern)
                app.router.register_resource(resource)
                if method.upper() == "GET":
                    resource.add_route("HEAD", handler, expect_handler=None)
                resource.add_route(method.upper(), handler, expect_handler=None)

            context["_http_accept_new_requests"] = True

            port = http_options.port
            host = http_options.host
            if port is True:
                raise ValueError("Bad value for http option port: {}".format(str(port)))

            # Configuration settings for keep-alive could use some refactoring

            keepalive_timeout_option = http_options.keepalive_timeout or http_options.keepalive_expiry
            keepalive_timeout = 0
            if keepalive_timeout_option is None or keepalive_timeout_option is False:
                keepalive_timeout_option = 0
            if keepalive_timeout_option is True:
                raise ValueError(
                    "Bad value for http option keepalive_timeout: {}".format(str(keepalive_timeout_option))
                )
            try:
                keepalive_timeout = int(keepalive_timeout_option) if keepalive_timeout_option is not None else 0
            except Exception:
                raise ValueError(
                    "Bad value for http option keepalive_timeout: {}".format(str(keepalive_timeout_option))
                ) from None

            tcp_keepalive = False
            if keepalive_timeout > 0:
                tcp_keepalive = True
            else:
                tcp_keepalive = False
                keepalive_timeout = 0

            max_keepalive_requests_option = http_options.max_keepalive_requests
            max_keepalive_requests = None
            if max_keepalive_requests_option is None or max_keepalive_requests_option is False:
                max_keepalive_requests_option = None
            if max_keepalive_requests_option is True:
                raise ValueError(
                    "Bad value for http option max_keepalive_requests: {}".format(str(max_keepalive_requests_option))
                )
            try:
                if max_keepalive_requests_option is not None:
                    max_keepalive_requests = int(max_keepalive_requests_option)
                if max_keepalive_requests == 0:
                    max_keepalive_requests = None
            except Exception:
                raise ValueError(
                    "Bad value for http option max_keepalive_requests: {}".format(str(max_keepalive_requests_option))
                ) from None
            if not tcp_keepalive and max_keepalive_requests:
                raise ValueError(
                    "HTTP keep-alive must be enabled to use http option max_keepalive_requests - a http.keepalive_timeout option value is required"
                ) from None

            max_keepalive_time_option = http_options.max_keepalive_time
            max_keepalive_time = None
            if max_keepalive_time_option is None or max_keepalive_time_option is False:
                max_keepalive_time_option = None
            if max_keepalive_time_option is True:
                raise ValueError(
                    "Bad value for http option max_keepalive_time: {}".format(str(max_keepalive_time_option))
                )
            try:
                if max_keepalive_time_option is not None:
                    max_keepalive_time = int(max_keepalive_time_option)
                if max_keepalive_time == 0:
                    max_keepalive_time = None
            except Exception:
                raise ValueError(
                    "Bad value for http option max_keepalive_time: {}".format(str(max_keepalive_time_option))
                ) from None
            if not tcp_keepalive and max_keepalive_time:
                raise ValueError(
                    "HTTP keep-alive must be enabled to use http option max_keepalive_time - a http.keepalive_timeout option value is required"
                ) from None

            reuse_port = True if http_options.reuse_port else False
            if reuse_port and platform.system() != "Linux":
                logger.warning(
                    "The http option reuse_port (socket.SO_REUSEPORT) can only enabled on Linux platforms - current "
                    f"platform is {platform.system()} - will revert option setting to not reuse ports"
                )
                reuse_port = False

            context["_http_tcp_keepalive"] = tcp_keepalive
            context["_http_keepalive_timeout"] = keepalive_timeout
            context["_http_max_keepalive_requests"] = max_keepalive_requests
            context["_http_max_keepalive_time"] = max_keepalive_time

            set_execution_context(
                {
                    "http_enabled": True,
                    "http_current_tasks": 0,
                    "http_total_tasks": 0,
                    "aiohttp_version": aiohttp_version,
                }
            )

            try:
                app.freeze()
                web_server = Server(
                    app._handle,
                    request_factory=app._make_request,
                    server_header=server_header,
                    access_log=access_log,
                    real_ip_header=real_ip_header,
                    real_ip_from=real_ip_from,
                    keepalive_timeout=keepalive_timeout,
                    tcp_keepalive=tcp_keepalive,
                )

                if reuse_port:
                    if not port:
                        logger.warning(
                            "The http option reuse_port (socket option SO_REUSEPORT) is enabled by default on Linux - "
                            "listening on random ports with SO_REUSEPORT is dangerous - please double check your intent"
                        )
                    elif str(port) in HttpTransport.server_port_mapping.values():
                        logger.warning(
                            "The http option reuse_port (socket option SO_REUSEPORT) is enabled by default on Linux - "
                            "different service classes should not use the same port ({})".format(port)
                        )

                if port:
                    HttpTransport.server_port_mapping[web_server] = str(port)

                server_task = loop.create_server(web_server, host, port, reuse_port=reuse_port)
                server = await server_task
            except OSError as e:
                context["_http_accept_new_requests"] = False
                error_message = re.sub(".*: ", "", e.strerror)
                logger.warning(
                    "unable to bind service [http] to http://{}:{}/".format(
                        "localhost" if host in ("0.0.0.0", "127.0.0.1") else host, port
                    ),
                    host=host,
                    port=port,
                    error_message=error_message,
                )

                try:
                    raise HttpException(str(e)).with_traceback(e.__traceback__) from None
                except Exception as exc:
                    exc.__traceback__ = e.__traceback__
                    raise

            if server.sockets:
                socket_address = server.sockets[0].getsockname()
                if socket_address:
                    port = int(socket_address[1])
                    HttpTransport.server_port_mapping[web_server] = str(port)
            context["_http_port"] = port

            stop_method = getattr(obj, "_stop_service", None)

            async def stop_service(*args: Any, **kwargs: Any) -> None:
                context["_http_tcp_keepalive"] = False

                server.close()
                await server.wait_closed()

                HttpTransport.server_port_mapping.pop(web_server, None)

                shutdown_sleep = 0
                if len(web_server.connections):
                    shutdown_sleep = 1
                    await asyncio.sleep(1)

                if not tcp_keepalive:
                    context["_http_accept_new_requests"] = False

                open_websockets = context.get("_http_open_websockets", [])[:]
                if open_websockets:
                    logger.info("closing websocket connections", connection_count=len(open_websockets))
                    tasks = []
                    for websocket in open_websockets:
                        try:
                            tasks.append(asyncio.ensure_future(websocket.close()))
                        except Exception:
                            pass
                    try:
                        await asyncio.wait_for(
                            asyncio.shield(asyncio.gather(*tasks, return_exceptions=True)), timeout=10
                        )
                        await asyncio.sleep(1)
                    except (Exception, asyncio.TimeoutError, asyncio.CancelledError):
                        pass
                    context["_http_open_websockets"] = []

                termination_grace_period_seconds = 30
                try:
                    termination_grace_period_seconds = int(
                        http_options.termination_grace_period_seconds or termination_grace_period_seconds
                    )
                except Exception:
                    pass

                log_wait_message = True if termination_grace_period_seconds >= 2 else False

                if tcp_keepalive and len(web_server.connections):
                    wait_start_time = time.time()

                    while wait_start_time + max(2, termination_grace_period_seconds) > time.time():
                        active_requests = context.get("_http_active_requests", set())
                        if not active_requests and not len(web_server.connections):
                            break

                        if log_wait_message:
                            log_wait_message = False
                            if len(web_server.connections) and len(web_server.connections) != len(active_requests):
                                logger.info(
                                    "awaiting keep-alive connections", connection_count=len(web_server.connections)
                                )
                            if active_requests:
                                logger.info(
                                    "awaiting requests to complete",
                                    request_count=len(active_requests),
                                    grace_period_seconds=termination_grace_period_seconds,
                                )

                        await asyncio.sleep(0.25)

                    termination_grace_period_seconds -= int(time.time() - wait_start_time)

                context["_http_accept_new_requests"] = False

                active_requests = context.get("_http_active_requests", set())
                if active_requests:
                    if log_wait_message:
                        logger.info(
                            "awaiting requests to complete",
                            request_count=len(active_requests),
                            grace_period_seconds=termination_grace_period_seconds,
                        )

                    try:
                        await asyncio.wait_for(
                            asyncio.shield(asyncio.gather(*active_requests, return_exceptions=True)),
                            timeout=max(2, termination_grace_period_seconds),
                        )
                        await asyncio.sleep(1)
                        active_requests = context.get("_http_active_requests", set())
                    except (Exception, asyncio.TimeoutError, asyncio.CancelledError):
                        active_requests = context.get("_http_active_requests", set())
                        if active_requests:
                            logger.warning(
                                "all requests did not gracefully finish execution",
                                remaining_request_count=len(active_requests),
                            )
                    context["_http_active_requests"] = set()

                if shutdown_sleep > 0:
                    await asyncio.sleep(shutdown_sleep)

                if len(web_server.connections):
                    logger.warning(
                        "forcefully closing open tcp connections", connection_count=len(web_server.connections)
                    )
                    await app.shutdown()
                    await asyncio.sleep(1)
                else:
                    await app.shutdown()

                if logger_handler:
                    response_logger = logging.getLogger("tomodachi.http.response")
                    response_logger.removeHandler(logger_handler)
                await app.cleanup()
                if stop_method:
                    await stop_method(*args, **kwargs)

            setattr(obj, "_stop_service", stop_service)

            for method, pattern, handler, route_context in context.get("_http_routes", []):
                for registry in getattr(obj, "discovery", []):
                    if getattr(registry, "add_http_endpoint", None):
                        await registry.add_http_endpoint(obj, host, port, method, pattern)

            listen_url = "http://{}:{}/".format("localhost" if host in ("0.0.0.0", "127.0.0.1") else host, port)
            logger.info("accepting http requests", listen_url=listen_url, listen_host=host, listen_port=port)

            if logger_handler:
                response_logger = logging.getLogger("tomodachi.http.response")
                response_logger._logger.propagate = False
                response_logger.info(
                    "accepting http requests", listen_url=listen_url, listen_host=host, listen_port=port
                )
                response_logger._logger.propagate = True

        return _start_server


async def resolve_response(
    value: Union[str, bytes, Dict, List, Tuple, web.Response, web.FileResponse, Response],
    request: Optional[web.Request] = None,
    context: Optional[Dict] = None,
    status_code: Optional[Union[str, int]] = None,
    default_content_type: Optional[str] = None,
    default_charset: Optional[str] = None,
) -> Union[web.Response, web.FileResponse]:
    return resolve_response_sync(
        value=value,
        request=request,
        context=context,
        status_code=status_code,
        default_content_type=default_content_type,
        default_charset=default_charset,
    )


def resolve_response_sync(
    value: Union[str, bytes, Dict, List, Tuple, web.Response, web.FileResponse, Response],
    request: Optional[web.Request] = None,
    context: Optional[Dict] = None,
    status_code: Optional[Union[str, int]] = None,
    default_content_type: Optional[str] = None,
    default_charset: Optional[str] = None,
) -> Union[web.Response, web.FileResponse]:
    if not context:
        context = {}
    if isinstance(value, Response):
        return value.get_aiohttp_response(
            context, default_content_type=default_content_type, default_charset=default_charset
        )
    if isinstance(value, web.FileResponse):
        return value

    status = (
        int(status_code)
        if status_code
        else (request is not None and request._cache.get("error_status_code", 200)) or 200
    )
    headers = None
    if isinstance(value, dict):
        body = value.get("body")
        _status: Optional[SupportsInt] = value.get("status")
        if _status and isinstance(_status, (int, str, bytes)):
            status = int(_status)
        _returned_headers = value.get("headers")
        if _returned_headers:
            returned_headers: Union[Mapping[str, Any], Iterable[Tuple[str, Any]]] = _returned_headers
            headers = CIMultiDict(returned_headers)
    elif isinstance(value, list) or isinstance(value, tuple):
        _status = value[0]
        if _status and isinstance(_status, (int, str, bytes)):
            status = int(_status)
        body = value[1]
        if len(value) > 2:
            returned_headers = value[2]
            headers = CIMultiDict(returned_headers)
    elif isinstance(value, web.Response):
        return value
    else:
        if value is None:
            value = ""  # type: ignore
        body = value

    return Response(
        body=body, status=status, headers=headers, content_type=default_content_type, charset=default_charset
    ).get_aiohttp_response(context)


async def get_http_response_status(
    value: Union[str, bytes, Dict, List, Tuple, web.Response, web.FileResponse, Response, Exception],
    request: Optional[web.Request] = None,
    verify_transport: bool = True,
) -> Optional[int]:
    if isinstance(value, Exception) or isinstance(value, web.HTTPException):
        status_code = int(getattr(value, "status", 500)) if value is not None else 500
        return status_code
    else:
        response = resolve_response_sync(value, request=request)
        status_code = int(response.status) if response is not None else 500
        if verify_transport and request is not None and request.transport is None:
            return 499
        else:
            return status_code


def get_http_response_status_sync(
    value: Any, request: Optional[web.Request] = None, verify_transport: bool = True
) -> Optional[int]:
    if isinstance(value, Exception) or isinstance(value, web.HTTPException):
        status_code = int(getattr(value, "status", 500)) if value is not None else 500
        return status_code

    if verify_transport and request is not None and hasattr(request, "transport") and request.transport is None:
        return 499

    if isinstance(value, Response) and value._status:
        return int(value._status)
    elif isinstance(value, (web.Response, web.FileResponse)) and value.status:
        return int(value.status)
    elif isinstance(value, dict):
        _status: Optional[SupportsInt] = value.get("status")
        if _status and isinstance(_status, (int, str, bytes)):
            return int(_status)
    elif isinstance(value, list) or isinstance(value, tuple):
        _status = value[0]
        if _status and isinstance(_status, (int, str, bytes)):
            return int(_status)
    elif value and hasattr(value, "_status") and getattr(value, "_status", None):
        return int(getattr(value, "_status"))
    elif value and hasattr(value, "status") and getattr(value, "status", None):
        return int(getattr(value, "status"))

    return int((request is not None and request._cache.get("error_status_code", 200)) or 200)


__http = HttpTransport.decorator(HttpTransport.request_handler)
__http_error = HttpTransport.decorator(HttpTransport.error_handler)
__http_static = HttpTransport.decorator(HttpTransport.static_request_handler)

__websocket = HttpTransport.decorator(HttpTransport.websocket_handler)
__ws = HttpTransport.decorator(HttpTransport.websocket_handler)


def http(
    method: Union[str, List[str], Tuple[str, ...]],
    url: str,
    *,
    ignore_logging: Union[bool, List[int], Tuple[int, ...]] = False,
    pre_handler_func: Optional[Callable] = None,
) -> Callable:
    return cast(Callable, __http(method, url, ignore_logging=ignore_logging, pre_handler_func=pre_handler_func))


def http_error(status_code: int) -> Callable:
    return cast(Callable, __http_error(status_code))


def http_static(
    path: str, base_url: str, *, ignore_logging: Union[bool, List[int], Tuple[int, ...]] = False
) -> Callable:
    return cast(Callable, __http_static(path, base_url, ignore_logging=ignore_logging))


def websocket(url: str) -> Callable:
    return cast(Callable, __websocket(url))


def ws(url: str) -> Callable:
    return cast(Callable, __ws(url))



from typing import ClassVar as _ClassVar
from typing import Mapping as _Mapping
from typing import Optional as _Optional
from typing import Union as _Union

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message

DESCRIPTOR: _descriptor.FileDescriptor

class Metadata(_message.Message):
    __slots__ = ["data_encoding", "message_uuid", "protocol_version", "timestamp", "topic"]
    DATA_ENCODING_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_UUID_FIELD_NUMBER: _ClassVar[int]
    PROTOCOL_VERSION_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    TOPIC_FIELD_NUMBER: _ClassVar[int]
    data_encoding: str
    message_uuid: str
    protocol_version: str
    timestamp: float
    topic: str
    def __init__(self, message_uuid: _Optional[str] = ..., protocol_version: _Optional[str] = ..., timestamp: _Optional[float] = ..., topic: _Optional[str] = ..., data_encoding: _Optional[str] = ...) -> None: ...

class SNSSQSMessage(_message.Message):
    __slots__ = ["data", "metadata", "service"]
    DATA_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    SERVICE_FIELD_NUMBER: _ClassVar[int]
    data: bytes
    metadata: Metadata
    service: Service
    def __init__(self, service: _Optional[_Union[Service, _Mapping]] = ..., metadata: _Optional[_Union[Metadata, _Mapping]] = ..., data: _Optional[bytes] = ...) -> None: ...

class Service(_message.Message):
    __slots__ = ["name", "uuid"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    UUID_FIELD_NUMBER: _ClassVar[int]
    name: str
    uuid: str
    def __init__(self, name: _Optional[str] = ..., uuid: _Optional[str] = ...) -> None: ...



from __future__ import annotations

import platform
from typing import Any, Dict, List, Mapping, Optional, Tuple, Type, TypeVar, Union, cast

from .interface import OptionsInterface

T = TypeVar("T")


def DEFAULT(cls: Type[T]) -> T:
    return cast(T, type("DEFAULT", (type(cls),), {"_default": True}))


class _HTTP(OptionsInterface):
    port: int
    host: Optional[str]
    reuse_port: bool
    content_type: str
    charset: str
    client_max_size: Union[str, int]
    termination_grace_period_seconds: int
    access_log: Union[bool, str]
    real_ip_from: Union[str, List[str]]
    real_ip_header: str
    keepalive_timeout: int
    keepalive_expiry: int
    max_keepalive_time: Optional[int]
    max_keepalive_requests: Optional[int]
    server_header: str

    _hierarchy: Tuple[str, ...] = ("http",)
    _legacy_fallback: Dict[str, Union[str, Tuple[str, ...]]] = {
        "max_buffer_size": "client_max_size",
        "max_upload_size": "client_max_size",
    }

    def __init__(
        self,
        *,
        port: int = 9700,
        host: Optional[str] = "0.0.0.0",
        reuse_port: bool = (True if platform.system() == "Linux" else False),
        content_type: str = "text/plain; charset=utf-8",
        charset: str = "utf-8",
        client_max_size: Union[str, int] = (1024**2) * 100,
        termination_grace_period_seconds: int = 30,
        access_log: Union[bool, str] = True,
        real_ip_from: Optional[Union[str, List[str]]] = None,
        real_ip_header: str = "X-Forwarded-For",
        keepalive_timeout: int = 0,
        keepalive_expiry: int = 0,
        max_keepalive_time: Optional[int] = None,
        max_keepalive_requests: Optional[int] = None,
        server_header: str = "tomodachi",
        **kwargs: Any,
    ):
        self.port = port
        self.host = host
        self.reuse_port = reuse_port
        self.content_type = content_type
        self.charset = charset
        self.client_max_size = client_max_size
        self.termination_grace_period_seconds = termination_grace_period_seconds
        self.access_log = access_log
        self.real_ip_from = real_ip_from if real_ip_from is not None and real_ip_from != "" else []
        self.real_ip_header = real_ip_header
        self.keepalive_timeout = keepalive_timeout
        self.keepalive_expiry = keepalive_expiry
        self.max_keepalive_time = max_keepalive_time
        self.max_keepalive_requests = max_keepalive_requests
        self.server_header = server_header

        self._load_keyword_options(**kwargs)


class _AWSSNSSQS(OptionsInterface):
    region_name: Optional[str]
    aws_access_key_id: Optional[str]
    aws_secret_access_key: Optional[str]
    topic_prefix: str
    queue_name_prefix: str
    sns_kms_master_key_id: Optional[str]
    sqs_kms_master_key_id: Optional[str]
    sqs_kms_data_key_reuse_period: Optional[int]
    queue_policy: Optional[str]
    wildcard_queue_policy: Optional[str]

    _hierarchy: Tuple[str, ...] = ("aws_sns_sqs",)
    _legacy_fallback: Dict[str, Union[str, Tuple[str, ...]]] = {
        "aws_region_name": "region_name",
        "secret_access_key": "aws_secret_access_key",
        "access_key_id": "aws_access_key_id",
        "endpoint_url": (".aws_endpoint_urls.sns", ".aws_endpoint_urls.sqs"),
        "aws_endpoint_url": (".aws_endpoint_urls.sns", ".aws_endpoint_urls.sqs"),
        "endpoint_urls.sns": ".aws_endpoint_urls.sns",
        "endpoint_urls.sqs": ".aws_endpoint_urls.sqs",
        "aws_sns_endpoint_url": ".aws_endpoint_urls.sns",
        "aws_sqs_endpoint_url": ".aws_endpoint_urls.sqs",
        "sns_endpoint_url": ".aws_endpoint_urls.sns",
        "sqs_endpoint_url": ".aws_endpoint_urls.sqs",
        "aws_kms_master_key_id": ("sns_kms_master_key_id", "sqs_kms_master_key_id"),
        "kms_master_key_id": ("sns_kms_master_key_id", "sqs_kms_master_key_id"),
        "aws_sns_kms_master_key_id": "sns_kms_master_key_id",
        "aws_sqs_kms_master_key_id": "sqs_kms_master_key_id",
        "aws_sqs_kms_data_key_reuse_period": "sqs_kms_data_key_reuse_period",
        "aws_kms_data_key_reuse_period": "sqs_kms_data_key_reuse_period",
        "kms_data_key_reuse_period": "sqs_kms_data_key_reuse_period",
    }

    def __init__(
        self,
        *,
        region_name: Optional[str] = None,
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        topic_prefix: str = "",
        queue_name_prefix: str = "",
        sns_kms_master_key_id: Optional[str] = None,
        sqs_kms_master_key_id: Optional[str] = None,
        sqs_kms_data_key_reuse_period: Optional[int] = None,
        queue_policy: Optional[str] = None,
        wildcard_queue_policy: Optional[str] = None,
        **kwargs: Any,
    ):
        self.region_name = region_name
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.topic_prefix = topic_prefix
        self.queue_name_prefix = queue_name_prefix
        self.sns_kms_master_key_id = sns_kms_master_key_id
        self.sqs_kms_master_key_id = sqs_kms_master_key_id
        self.sqs_kms_data_key_reuse_period = sqs_kms_data_key_reuse_period
        self.queue_policy = queue_policy
        self.wildcard_queue_policy = wildcard_queue_policy

        self._load_keyword_options(**kwargs)


class _AWSEndpointURLs(OptionsInterface):
    sns: Optional[str]
    sqs: Optional[str]

    _hierarchy: Tuple[str, ...] = ("aws_endpoint_urls",)

    def __init__(
        self,
        *,
        sns: Optional[str] = None,
        sqs: Optional[str] = None,
        **kwargs: Any,
    ):
        self.sns = sns
        self.sqs = sqs

        self._load_keyword_options(**kwargs)


class _AMQP_QOS(OptionsInterface):
    queue_prefetch_count: int
    global_prefetch_count: int

    _hierarchy: Tuple[str, ...] = ("amqp", "qos")

    def __init__(
        self,
        *,
        queue_prefetch_count: int = 100,
        global_prefetch_count: int = 400,
        **kwargs: Any,
    ):
        self.queue_prefetch_count = queue_prefetch_count
        self.global_prefetch_count = global_prefetch_count

        self._load_keyword_options(**kwargs)


class _AMQP(OptionsInterface):
    class QOS(_AMQP_QOS):
        pass

    host: str
    port: int
    login: str
    password: str
    exchange_name: str
    routing_key_prefix: str
    queue_name_prefix: str
    virtualhost: str
    ssl: bool
    heartbeat: int
    queue_ttl: int
    qos: QOS

    _hierarchy: Tuple[str, ...] = ("amqp",)

    def __init__(
        self,
        *,
        host: str = "127.0.0.1",
        port: int = 5672,
        login: str = "guest",
        password: str = "guest",
        exchange_name: str = "amq.topic",
        routing_key_prefix: str = "",
        queue_name_prefix: str = "",
        virtualhost: str = "/",
        ssl: bool = False,
        heartbeat: int = 60,
        queue_ttl: int = 86400,
        qos: Union[Mapping[str, Any], QOS] = DEFAULT(QOS),
        **kwargs: Any,
    ):
        self.host = host
        self.port = port
        self.login = login
        self.password = password
        self.exchange_name = exchange_name
        self.routing_key_prefix = routing_key_prefix
        self.queue_name_prefix = queue_name_prefix
        self.virtualhost = virtualhost
        self.ssl = ssl
        self.heartbeat = heartbeat
        self.queue_ttl = queue_ttl

        input_: Tuple[Tuple[str, Union[Mapping[str, Any], OptionsInterface], type], ...] = (("qos", qos, self.QOS),)
        self._load_initial_input(input_)
        self._load_keyword_options(**kwargs)


class _Watcher(OptionsInterface):
    ignored_dirs: List[str]
    watched_file_endings: List[str]

    _hierarchy: Tuple[str, ...] = ("watcher",)
    __slots__: Tuple[str, ...] = ("ignored_dirs", "watched_file_endings")

    def __init__(
        self,
        *,
        ignored_dirs: Optional[List[str]] = None,
        watched_file_endings: Optional[List[str]] = None,
        **kwargs: Any,
    ):
        self.ignored_dirs = ignored_dirs if ignored_dirs is not None else []
        self.watched_file_endings = watched_file_endings if watched_file_endings is not None else []

        self._load_keyword_options(**kwargs)


class Options(OptionsInterface):
    class HTTP(_HTTP):
        pass

    class AWSSNSSQS(_AWSSNSSQS):
        pass

    class AWSEndpointURLs(_AWSEndpointURLs):
        pass

    class AMQP(_AMQP):
        class QOS(_AMQP_QOS):
            pass

    class Watcher(_Watcher):
        pass

    http: HTTP
    aws_sns_sqs: AWSSNSSQS
    aws_endpoint_urls: AWSEndpointURLs
    amqp: AMQP
    watcher: Watcher

    _hierarchy: Tuple[str, ...] = ()
    _legacy_fallback: Dict[str, Union[str, Tuple[str, ...]]] = {
        "aws.region_name": "aws_sns_sqs.region_name",
        "aws.aws_region_name": "aws_sns_sqs.region_name",
        "aws.secret_access_key": "aws_sns_sqs.aws_secret_access_key",
        "aws.aws_secret_access_key": "aws_sns_sqs.aws_secret_access_key",
        "aws.access_key_id": "aws_sns_sqs.aws_access_key_id",
        "aws.aws_access_key_id": "aws_sns_sqs.aws_access_key_id",
        "aws.endpoint_url": ("aws_endpoint_urls.sns", "aws_endpoint_urls.sqs"),
        "aws.aws_endpoint_url": ("aws_endpoint_urls.sns", "aws_endpoint_urls.sqs"),
        "aws.endpoint_urls.sns": "aws_endpoint_urls.sns",
        "aws.endpoint_urls.sqs": "aws_endpoint_urls.sqs",
        "aws.aws_sns_endpoint_url": "aws_endpoint_urls.sns",
        "aws.aws_sqs_endpoint_url": "aws_endpoint_urls.sqs",
        "aws.sns_endpoint_url": "aws_endpoint_urls.sns",
        "aws.sqs_endpoint_url": "aws_endpoint_urls.sqs",
        "aws.topic_prefix": "aws_sns_sqs.topic_prefix",
        "aws.queue_name_prefix": "aws_sns_sqs.queue_name_prefix",
        "aws.aws_kms_master_key_id": ("aws_sns_sqs.sns_kms_master_key_id", "aws_sns_sqs.sqs_kms_master_key_id"),
        "aws.kms_master_key_id": ("aws_sns_sqs.sns_kms_master_key_id", "aws_sns_sqs.sqs_kms_master_key_id"),
        "aws.aws_sns_kms_master_key_id": "aws_sns_sqs.sns_kms_master_key_id",
        "aws.aws_sqs_kms_master_key_id": "aws_sns_sqs.sqs_kms_master_key_id",
        "aws.sns_kms_master_key_id": "aws_sns_sqs.sns_kms_master_key_id",
        "aws.sqs_kms_master_key_id": "aws_sns_sqs.sqs_kms_master_key_id",
        "aws.aws_sqs_kms_data_key_reuse_period": "aws_sns_sqs.sqs_kms_data_key_reuse_period",
        "aws.aws_kms_data_key_reuse_period": "aws_sns_sqs.sqs_kms_data_key_reuse_period",
        "aws.kms_data_key_reuse_period": "aws_sns_sqs.sqs_kms_data_key_reuse_period",
        "aws.queue_policy": "aws_sns_sqs.queue_policy",
        "aws.wildcard_queue_policy": "aws_sns_sqs.wildcard_queue_policy",
    }

    def __init__(
        self,
        *,
        http: Union[Mapping[str, Any], HTTP] = DEFAULT(HTTP),
        aws_sns_sqs: Union[Mapping[str, Any], AWSSNSSQS] = DEFAULT(AWSSNSSQS),
        aws_endpoint_urls: Union[Mapping[str, Any], AWSEndpointURLs] = DEFAULT(AWSEndpointURLs),
        amqp: Union[Mapping[str, Any], AMQP] = DEFAULT(AMQP),
        watcher: Union[Mapping[str, Any], Watcher] = DEFAULT(Watcher),
        **kwargs: Any,
    ):
        input_: Tuple[Tuple[str, Union[Mapping[str, Any], OptionsInterface], type], ...] = (
            ("http", http, self.HTTP),
            ("aws_sns_sqs", aws_sns_sqs, self.AWSSNSSQS),
            ("aws_endpoint_urls", aws_endpoint_urls, self.AWSEndpointURLs),
            ("amqp", amqp, self.AMQP),
            ("watcher", watcher, self.Watcher),
        )

        self._load_initial_input(input_)
        self._load_keyword_options(**kwargs)



# flake8: noqa
import contextvars
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

from tomodachi import cli as cli
from tomodachi import container as container
from tomodachi import discovery as discovery
from tomodachi import envelope as envelope
from tomodachi import helpers as helpers
from tomodachi import importer as importer
from tomodachi import invoker as invoker
from tomodachi import launcher as launcher
from tomodachi import logging as logging
from tomodachi import opentelemetry as opentelemetry
from tomodachi import options as options
from tomodachi import transport as transport
from tomodachi import watcher as watcher
from tomodachi.__version__ import __build_time__ as __build_time__
from tomodachi.__version__ import __version__ as __version__
from tomodachi.__version__ import __version_info__ as __version_info__
from tomodachi.helpers.aiobotocore_connector import ClientConnector as _AiobotocoreClientConnector
from tomodachi.helpers.aiobotocore_connector import connector as _aiobotocore_client_connector
from tomodachi.helpers.execution_context import clear_execution_context as _clear_execution_context
from tomodachi.helpers.execution_context import clear_services as _clear_services
from tomodachi.helpers.execution_context import decrease_execution_context_value as decrease_execution_context_value
from tomodachi.helpers.execution_context import get_execution_context as get_execution_context
from tomodachi.helpers.execution_context import get_instance as get_instance
from tomodachi.helpers.execution_context import get_service as get_service
from tomodachi.helpers.execution_context import get_services as get_services
from tomodachi.helpers.execution_context import increase_execution_context_value as increase_execution_context_value
from tomodachi.helpers.execution_context import set_execution_context as set_execution_context
from tomodachi.helpers.execution_context import set_service as _set_service
from tomodachi.helpers.execution_context import unset_service as _unset_service
from tomodachi.invoker import decorator as decorator
from tomodachi.logging import Logger as Logger
from tomodachi.logging import LoggerProtocol as LoggerProtocol
from tomodachi.logging import get_logger as get_logger
from tomodachi.options import Options as Options
from tomodachi.options import OptionsInterface as OptionsInterface
from tomodachi.transport.amqp import amqp as amqp
from tomodachi.transport.amqp import amqp_publish as amqp_publish
from tomodachi.transport.aws_sns_sqs import aws_sns_sqs as aws_sns_sqs
from tomodachi.transport.aws_sns_sqs import aws_sns_sqs_publish as aws_sns_sqs_publish
from tomodachi.transport.aws_sns_sqs import awssnssqs as awssnssqs
from tomodachi.transport.aws_sns_sqs import awssnssqs_publish as awssnssqs_publish
from tomodachi.transport.http import HttpException as HttpException
from tomodachi.transport.http import Response as _HttpResponse
from tomodachi.transport.http import get_forwarded_remote_ip as get_forwarded_remote_ip
from tomodachi.transport.http import get_http_response_status as get_http_response_status
from tomodachi.transport.http import get_http_response_status_sync as get_http_response_status_sync
from tomodachi.transport.http import http as http
from tomodachi.transport.http import http_error as http_error
from tomodachi.transport.http import http_static as http_static
from tomodachi.transport.http import websocket as websocket
from tomodachi.transport.http import ws as ws
from tomodachi.transport.schedule import daily as daily
from tomodachi.transport.schedule import every_second as every_second
from tomodachi.transport.schedule import heartbeat as heartbeat
from tomodachi.transport.schedule import hourly as hourly
from tomodachi.transport.schedule import minutely as minutely
from tomodachi.transport.schedule import monthly as monthly
from tomodachi.transport.schedule import schedule as schedule

AiobotocoreClientConnector = _AiobotocoreClientConnector
aiobotocore_client_connector = _aiobotocore_client_connector
HttpResponse = _HttpResponse

__author__: str = ...
__email__: str = ...

CLASS_ATTRIBUTE: str = ...
TOMODACHI_CLASSES: List[Type] = ...
DEFAULT_SERVICE_EXIT_CODE: int = ...
SERVICE_EXIT_CODE: int = ...

def get_contextvar(key: str) -> contextvars.ContextVar: ...
def context(key: str) -> Any: ...

class TomodachiServiceMeta(type):
    def __new__(
        cls: Type[TomodachiServiceMeta], name: str, bases: Tuple[type, ...], attributedict: Dict
    ) -> TomodachiServiceMeta: ...

class Service(metaclass=TomodachiServiceMeta):
    _tomodachi_class_is_service_class: bool = ...
    name: str = ...
    uuid: str = ...
    options: Options
    logger: LoggerProtocol
    log: Callable = ...

def service(cls: Type[object]) -> Type[TomodachiServiceMeta]: ...
def exit(exit_code: Optional[int] = None) -> None: ...
def run(app: Optional[Union[str, List[str], Tuple[str]]] = None, *args: str, **kwargs: Optional[str]) -> None: ...



from __future__ import annotations

import datetime
import json
import logging
import os
import sys
import warnings
from contextvars import ContextVar
from functools import wraps
from io import StringIO
from logging import CRITICAL, DEBUG, ERROR, FATAL, INFO, NOTSET, WARN, WARNING
from types import ModuleType
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    KeysView,
    Literal,
    Optional,
    Protocol,
    Sequence,
    TextIO,
    Tuple,
    Type,
    Union,
    cast,
    overload,
)

import structlog
from structlog._log_levels import _LEVEL_TO_NAME, _NAME_TO_LEVEL
from structlog.exceptions import DropEvent

from tomodachi.helpers.colors import NO_COLOR

if TYPE_CHECKING:  # pragma: no cover
    try:
        from structlog.typing import Context, EventDict, ExcInfo, Processor, WrappedLogger
    except (ImportError, ModuleNotFoundError):
        from structlog.types import Context, EventDict, ExcInfo, Processor, WrappedLogger

MAX_STACKTRACE_DEPTH = 50
RENAME_KEYS: Sequence[Tuple[str, str]] = (("event", "message"), ("event_", "event"), ("class_", "class"))
EXCEPTION_KEYS: Sequence[str] = ("exception", "exc", "error", "message")
CONSOLE_QUOTE_KEYS = ("tb_location", "error_location", "tb_filename", "co_filename", "co_location", "error_filename")
TOMODACHI_LOGGER_TYPE: Literal["json", "console", "no_color_console", "custom", "python", "disabled"] = "console"
TOMODACHI_CUSTOM_LOGGER: Optional[Union[str, ModuleType, type, object]] = None

STD_LOGGER_FIELDS = set(
    [
        "name",
        "levelno",
        "levelname",
        "pathname",
        "filename",
        "module",
        "lineno",
        "funcName",
        "created",
        "asctime",
        "msecs",
        "relativeCreated",
        "thread",
        "threadName",
        "taskName",
        "process",
        "message",
        "args",
        "exc_text",
        "stack_info",
        "msg",
        "processName",
        "exc_info",
    ]
)

_context: ContextVar[Union[LoggerContext, Dict]] = ContextVar("tomodachi.logging._context", default={})
_loggers: ContextVar[Dict] = ContextVar("tomodachi.logging._loggers", default={})


class LogProcessorTimestamp:
    __slots__ = ("key",)

    def __init__(
        self,
        key: str = "timestamp",
    ) -> None:
        self.key = key

    def __call__(self, logger: WrappedLogger, method_name: str, event_dict: EventDict) -> EventDict:
        if self.key not in event_dict:
            event_dict[self.key] = (
                datetime.datetime.now(datetime.timezone.utc).isoformat(timespec="microseconds").replace("+00:00", "Z")
            )
        return event_dict


class AddMissingDictKey:
    __slots__ = ("key",)

    def __init__(
        self,
        key: str,
    ) -> None:
        self.key = key

    def __call__(self, logger: WrappedLogger, method_name: str, event_dict: EventDict) -> EventDict:
        if self.key not in event_dict:
            event_dict[self.key] = ""
        return event_dict


class RemoveDictKey:
    __slots__ = ("key",)

    def __init__(
        self,
        key: str,
    ) -> None:
        self.key = key

    def __call__(self, logger: WrappedLogger, method_name: str, event_dict: EventDict) -> EventDict:
        if self.key in event_dict:
            event_dict.pop(self.key)
        return event_dict


class LinkQuoteStrings:
    __slots__ = ("keys",)

    def __init__(
        self,
        keys: Iterable[str],
    ) -> None:
        self.keys = keys

    def __call__(self, logger: WrappedLogger, method_name: str, event_dict: EventDict) -> EventDict:
        return {k: f"<{v}>" if k in self.keys and v and not v.startswith("<") else v for k, v in event_dict.items()}


class RenameKeys:
    __slots__ = ("pairs",)

    def __init__(
        self,
        pairs: Union[Sequence[Tuple[str, str]], Dict[str, str]],
    ) -> None:
        if not isinstance(pairs, dict):
            pairs = {k: v for k, v in pairs}
        self.pairs = pairs

    def __call__(self, logger: WrappedLogger, method_name: str, event_dict: EventDict) -> EventDict:
        return {(self.pairs.get(k, k)): v for k, v in event_dict.items()}


_ordered_items = structlog.processors._items_sorter(
    sort_keys=False, key_order=("timestamp", "logger", "level", "message"), drop_missing=True
)


def serializer_func(event_dict: EventDict, **dumps_kw: Any) -> str:
    return json.dumps(dict(_ordered_items(event_dict)), **dumps_kw)


def merge_contextvars(logger: WrappedLogger, method_name: str, event_dict: EventDict) -> EventDict:
    ctx = structlog.contextvars._CONTEXT_VARS.copy()

    for k, v in ctx.items():
        if k.startswith(structlog.contextvars.STRUCTLOG_KEY_PREFIX) and v.get() is not Ellipsis:
            event_dict[k[structlog.contextvars.STRUCTLOG_KEY_PREFIX_LEN :]] = v.get()

    return event_dict


def add_exception_info(logger: WrappedLogger, method_name: str, event_dict: EventDict) -> EventDict:
    exception = None
    for key in EXCEPTION_KEYS:
        if key in event_dict and isinstance(event_dict.get(key), BaseException):
            exception = event_dict.get(key)
            break

    exc_info = event_dict.get("exc_info")

    if method_name == "error" and exc_info is True:
        method_name = "exception"

    if exception and isinstance(exception, BaseException):
        pass
    elif (
        not isinstance(exception, BaseException)
        and isinstance(exc_info, (tuple, list))
        and len(exc_info) == 3
        and isinstance(exc_info[1], BaseException)
    ):
        exception = exc_info[1]
    elif method_name == "exception":
        exc_info = sys.exc_info()
        exception = exc_info[1] if exc_info and isinstance(exc_info[1], BaseException) else None

    if not exception or not isinstance(exception, BaseException):
        return event_dict

    for key in EXCEPTION_KEYS:
        if key in event_dict and exception == event_dict.get(key):
            event_dict.pop(key)

    exc_info = (type(exception), exception, exception.__traceback__)

    event_dict["exception"] = exception
    event_dict["exc_info"] = exc_info

    if "exc_type" not in event_dict:
        event_dict["exc_type"] = (
            type(exception).__name__ if hasattr(type(exception), "__name__") else str(type(exception))
        )
    if "exc_message" not in event_dict:
        event_dict["exc_message"] = str(exception)

    tb = exception.__traceback__
    while tb and tb.tb_next:
        tb = tb.tb_next

    if "tb_module_name" not in event_dict:
        event_dict["tb_module_name"] = tb.tb_frame.f_globals.get("__name__", "<unknown>") if tb else "<unknown>"
    if "tb_function_name" not in event_dict:
        event_dict["tb_function_name"] = (
            getattr(tb.tb_frame.f_code, "co_qualname", tb.tb_frame.f_code.co_name) if tb else "<unknown>"
        )
    if "tb_location" not in event_dict:
        co_filename = tb.tb_frame.f_code.co_filename if tb else "<unknown>"
        if tb:
            cwd = os.getcwd()
            if cwd.rstrip("/") and co_filename.startswith(cwd):
                co_filename = "./" + co_filename[len(cwd) :].lstrip("/")

        event_dict["tb_location"] = f"{co_filename}:{tb.tb_lineno}" if tb else "<unknown>"

    return event_dict


def add_stacktrace_info(logger: WrappedLogger, method_name: str, event_dict: EventDict) -> EventDict:
    exception = event_dict.get("exception")

    if not exception or not isinstance(exception, BaseException):
        return event_dict

    if "stacktrace" in event_dict:
        return event_dict

    stacktrace = []
    cwd = os.getcwd()

    tb = exception.__traceback__
    while tb:
        co_filename = tb.tb_frame.f_code.co_filename or "<unknown>"
        if cwd.rstrip("/") and co_filename.startswith(cwd):
            co_filename = "./" + co_filename[len(cwd) :].lstrip("/")

        stacktrace.append(
            {
                "module_name": tb.tb_frame.f_globals.get("__name__", "<unknown>"),
                "function_name": getattr(tb.tb_frame.f_code, "co_qualname", tb.tb_frame.f_code.co_name),
                "location": f"{co_filename}:{tb.tb_lineno}",
            }
        )
        tb = tb.tb_next

    if len(stacktrace) > MAX_STACKTRACE_DEPTH:
        stacktrace = stacktrace[len(stacktrace) - MAX_STACKTRACE_DEPTH :]

    event_dict["stacktrace"] = stacktrace
    return event_dict


def remove_ellipsis_values(logger: WrappedLogger, method_name: str, event_dict: EventDict) -> EventDict:
    return {k: v for k, v in event_dict.items() if v is not Ellipsis}


def to_logger_args_kwargs(
    logger: WrappedLogger, method_name: str, event_dict: EventDict
) -> Tuple[Tuple, Dict[str, Any]]:
    return ((event_dict.get("event") or "",), {"extra": {"logger.context": event_dict}})


def to_custom_logger(
    logger: WrappedLogger, method_name: str, event_dict: EventDict
) -> Union[str, Tuple[Tuple, Dict[str, Any]]]:
    if isinstance(logger, logging.Logger):
        if event_dict.get("extra", {}).get("_processed"):
            return str(event_dict["extra"].get("logger.context", {}).get("event", ""))
        args, kwargs = to_logger_args_kwargs(logger, method_name, event_dict)
        kwargs["extra"]["_processed"] = True
        getattr(logger, method_name)(args, kwargs)
    else:
        event = event_dict.pop("event", "") or ""
        getattr(logger, method_name)(event, **event_dict)

    raise DropEvent


def disabled_logger_processor(logger: WrappedLogger, method_name: str, event_dict: EventDict) -> EventDict:
    raise DropEvent


class _PythonLoggingLoggerFormatter(logging.Formatter):
    style: Union[logging.PercentStyle, logging.StrFormatStyle, logging.StringTemplateStyle]
    fmt: str

    def __init__(
        self,
        fmt: Optional[str] = None,
        datefmt: Optional[str] = None,
        style: str = "%",
        validate: bool = True,
        *,
        defaults: Optional[Dict[str, Any]] = None,
    ):
        if style not in logging._STYLES:
            raise ValueError("Style must be one of: %s" % ",".join(logging._STYLES.keys()))
        style_cls = cast(
            Union[Type[logging.PercentStyle], Type[logging.StrFormatStyle], Type[logging.StringTemplateStyle]],
            logging._STYLES[style][0],
        )
        self.style = style_cls(fmt or "")
        if defaults and sys.version_info.major == 3 and sys.version_info.minor >= 10:
            setattr(self.style, "_defaults", defaults)

        if validate:
            self.style.validate()

        self.fmt = self.style._fmt
        self.datefmt = datefmt

    @property
    def _style(self) -> Union[logging.PercentStyle, logging.StrFormatStyle, logging.StringTemplateStyle]:  # type: ignore
        if self.style._fmt != self.fmt:
            self.fmt = self.style._fmt
        return self.style

    @property  # type: ignore
    def _fmt(self) -> str:
        if self.style._fmt != self.fmt:
            self.fmt = self._style._fmt
        return self.fmt

    @_fmt.setter
    def _fmt(self, fmt: str) -> None:
        self.fmt = fmt
        self.style._fmt = fmt

    def formatTime(self, record: logging.LogRecord, datefmt: Optional[str] = None) -> str:
        return (
            datetime.datetime.fromtimestamp(record.created, tz=datetime.timezone.utc)
            .isoformat(timespec="microseconds")
            .replace("+00:00", "Z")
        )

    def __repr__(self) -> str:
        return "PythonLoggingFormatter"


class _StdLoggingFormatter(logging.Formatter):
    _logger_type: Literal["json", "console", "no_color_console", "custom", "python", "disabled"]

    def __init__(
        self,
        logger_type: Literal[
            "json", "console", "no_color_console", "custom", "python", "disabled"
        ] = TOMODACHI_LOGGER_TYPE,
        name: str = "",
        *a: Any,
        **kw: Any,
    ) -> None:
        self._logger_type = logger_type
        self._name = name

    def format(self, record: logging.LogRecord) -> str:
        if "logger.context" in record.__dict__:
            kw: Dict[str, Any] = record.__dict__["logger.context"]
        else:
            extra_keys = set(record.__dict__.keys() - STD_LOGGER_FIELDS)
            if (
                record.__dict__.get("args")
                and isinstance(record.__dict__.get("args"), dict)
                and record.__dict__.get("args", {}).get("extra", {})
            ):
                kw = {
                    "extra": {
                        **{k: v for k, v in record.__dict__.items() if k in extra_keys},
                        **record.__dict__.get("args", {}).get("extra", {}),
                    }
                }
            else:
                kw = {"extra": {k: v for k, v in record.__dict__.items() if k in extra_keys}}
            if record.exc_info and "exception" not in kw:
                kw["exception"] = record.exc_info[1]
                kw["exc_info"] = True

        if "timestamp" not in kw:
            kw["timestamp"] = self.formatTime(record)
        if "event" not in kw:
            kw["event"] = record.getMessage()

        kw["logger"] = record.name

        if "extra" in kw and isinstance(kw["extra"], dict) and len(kw["extra"]) == 0:
            kw.pop("extra")

        method_name = _LEVEL_TO_NAME[record.levelno]
        if method_name == "error" and kw and kw.get("exc_info") is True:
            method_name = "exception"

        logger = _get_logger(record.name, logger_type=self._logger_type)

        try:
            args, _ = logger._process_event(
                method_name,
                None,
                kw,
            )
            return cast(str, args[0])
        except DropEvent:
            if record.module == "_base":
                raise
            return ""

    def formatTime(self, record: logging.LogRecord, datefmt: Optional[str] = None) -> str:
        return (
            datetime.datetime.fromtimestamp(record.created, tz=datetime.timezone.utc)
            .isoformat(timespec="microseconds")
            .replace("+00:00", "Z")
        )

    def __repr__(self) -> str:
        name = self._name or self.__class__.__name__
        return name


class StderrHandler(logging.StreamHandler):
    def __init__(self, level: int = NOTSET) -> None:
        logging.Handler.__init__(self, level)

    @property
    def stream(self) -> TextIO:  # type: ignore
        return sys.stderr


DEFAULT_FORMAT = "%(asctime)s [%(levelname)-9s] %(message)-30s [%(name)s]"
_default_fmt = DEFAULT_FORMAT
try:
    if len(logging.root.handlers) and logging.root.handlers[0].formatter and logging.root.handlers[0].formatter._fmt:
        _default_fmt = logging.root.handlers[0].formatter._fmt
except AttributeError:
    pass

PythonLoggingFormatter = _PythonLoggingLoggerFormatter(fmt=_default_fmt)

ConsoleFormatter = _StdLoggingFormatter(logger_type="console", name="ConsoleFormatter")
NoColorConsoleFormatter = _StdLoggingFormatter(logger_type="no_color_console", name="NoColorConsoleFormatter")
JSONFormatter = _StdLoggingFormatter(logger_type="json", name="JSONFormatter")
DisabledFormatter = _StdLoggingFormatter(logger_type="disabled", name="DisabledFormatter")
CustomLoggerFormatter = _StdLoggingFormatter(logger_type="custom", name="CustomLoggerFormatter")

DefaultHandler = DefaultRootLoggerHandler = _defaultHandler = StderrHandler()


@overload
def set_default_formatter(
    *, logger_type: Literal["json", "console", "no_color_console", "custom", "python", "disabled"]
) -> None: ...


@overload
def set_default_formatter(
    logger_type: Literal["json", "console", "no_color_console", "custom", "python", "disabled"],
    /,
) -> None: ...


@overload
def set_default_formatter(*, formatter: logging.Formatter) -> None: ...


@overload
def set_default_formatter(formatter: logging.Formatter, /) -> None: ...


@overload
def set_default_formatter(_arg: Literal[None] = None, /) -> None: ...


def set_default_formatter(
    _arg: Optional[
        Union[logging.Formatter, Literal["json", "console", "no_color_console", "custom", "python", "disabled"]]
    ] = None,
    /,
    *,
    logger_type: Optional[Literal["json", "console", "no_color_console", "custom", "python", "disabled"]] = None,
    formatter: Optional[logging.Formatter] = None,
) -> None:
    if _arg is not None and (logger_type is not None or formatter is not None):
        raise TypeError("Invalid combination of arguments given to set_default_formatter()")
    if logger_type is not None and formatter is not None:
        raise TypeError("Invalid combination of arguments given to set_default_formatter()")

    global TOMODACHI_LOGGER_TYPE

    if _arg is None and logger_type is None and formatter is None:
        logger_type = TOMODACHI_LOGGER_TYPE

    if _arg is not None:
        if isinstance(_arg, logging.Formatter):
            formatter = _arg
        elif isinstance(_arg, str):
            logger_type = _arg
        else:
            raise TypeError("Invalid logger type or formatter given")

    if logger_type is not None:
        formatter = (
            JSONFormatter
            if logger_type == "json"
            else (
                ConsoleFormatter
                if logger_type == "console"
                else (
                    NoColorConsoleFormatter
                    if logger_type == "no_color_console"
                    else (
                        CustomLoggerFormatter
                        if logger_type == "custom"
                        else (
                            DisabledFormatter
                            if logger_type == "disabled"
                            else PythonLoggingFormatter if logger_type == "python" else None
                        )
                    )
                )
            )
        )

        if not formatter:
            raise Exception(
                "Invalid logger type: '{}' (exected 'console', 'json', 'custom', 'python' or 'disabled')".format(
                    logger_type
                )
            )

    if not formatter:
        raise Exception("Argument formatter missing (expected instance of logging.Formatter)')")

    if isinstance(formatter, _StdLoggingFormatter):
        TOMODACHI_LOGGER_TYPE = formatter._logger_type
    else:
        TOMODACHI_LOGGER_TYPE = "python"

    DefaultHandler.setFormatter(formatter)


set_default_formatter()


def _patch_structlog_callsite_parameter_adder_processor() -> None:
    # patch structlog to ignore tomodachi.logger when adding callsite information
    cls_init = structlog.processors.CallsiteParameterAdder.__init__

    if getattr(cls_init, "__wrapped__", None):
        return

    @wraps(cls_init)
    def __init__(self: structlog.processors.CallsiteParameterAdder, *a: Any, **kw: Any) -> None:
        cls_init(self, *a, **kw)
        if "tomodachi.logging" not in self._additional_ignores:
            self._additional_ignores.append("tomodachi.logging")

    setattr(structlog.processors.CallsiteParameterAdder, "__init__", __init__)


_patch_structlog_callsite_parameter_adder_processor()


def set_custom_logger_factory(
    logger_factory: Optional[Union[str, ModuleType, type, object]] = TOMODACHI_CUSTOM_LOGGER,
) -> None:
    if logger_factory is not None and logger_factory:
        logger_factory_ = _CustomLogger.get_logger_factory(logger_factory)
        if logger_factory_ is sys.modules[__name__]:
            # cannot set custom logger factory to tomodachi.logger itself
            logger_factory = None
            set_default_formatter("console")
        else:
            # patch structlog logger to ignore tomodachi.logger when adding callsite information
            for processor in structlog.get_config().get("processors", []):
                if isinstance(processor, structlog.processors.CallsiteParameterAdder):
                    if "tomodachi.logging" not in processor._additional_ignores:
                        processor._additional_ignores.append("tomodachi.logging")

    global TOMODACHI_CUSTOM_LOGGER
    TOMODACHI_CUSTOM_LOGGER = logger_factory


set_custom_logger_factory()


class LoggerContext(dict):
    def __init__(self, *a: Any, **kw: Any) -> None:
        if a:
            self.__data = {**a[0], **kw}
        else:
            self.__data = {**kw}

        name = self._data.get("logger")
        if name and kw:
            ctx = _loggers.get().get(name)
            if ctx != self:
                _loggers.set({**_loggers.get(), **{self._data["logger"]: self}})
                if name and name == _context.get().get("logger"):
                    _context.set(self)

    def __str__(self) -> str:
        return str(self._data)

    def __repr__(self) -> str:
        return str(self._data)

    def __getitem__(self, item: str) -> Any:
        value = self._data.get(item)
        return value

    def __setitem__(self, item: str, value: Any) -> None:
        self.__data[item] = value

    def __delitem__(self, item: str) -> None:
        del self._data[item]

    def __getattr__(self, item: str) -> Any:
        return self.__getitem__(item)

    def __bool__(self) -> bool:
        if self._data:
            return True
        else:
            return False

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self) -> Any:
        return self._data.__iter__()

    def __contains__(self, item: Any) -> Any:
        return self._data.__contains__(item)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, LoggerContext):
            return bool(self._data == other._data)
        elif isinstance(other, dict):
            return bool(self._data == other)
        else:
            return False

    def __ne__(self, other: Any) -> bool:
        return not self == other

    def keys(self) -> KeysView:  # type: ignore
        return self._data.keys()

    def get(self, key: str, default: Any = None) -> Any:
        return self._data.get(key, default)

    def copy(self) -> dict:
        return {**self._data}

    def update(self, *a: Any, **kw: Any) -> None:
        self._data.update(*a, **kw)

    def pop(self, item: str, *default: Any) -> Any:
        if not default:
            return self._data.pop(item)
        return self._data.pop(item, default[0])

    def clear(self) -> None:
        self._data.clear()

    @property
    def _data(self) -> Dict[str, Any]:
        return self.__data


class _CustomLogger:
    _logger_factory_cache: Dict[str, Any] = {}

    @classmethod
    def get_logger_factory(cls, custom_logger_factory: Optional[Union[str, ModuleType, type, object]]) -> Any:
        if not custom_logger_factory:
            raise Exception("Custom logger not set")

        logger_factory: Any

        if isinstance(custom_logger_factory, str):
            if custom_logger_factory in cls._logger_factory_cache:
                return cls._logger_factory_cache[custom_logger_factory]

            if custom_logger_factory in sys.modules:
                module_name = custom_logger_factory
                logger_factory = sys.modules[module_name]
            elif "." in custom_logger_factory:
                module_name, module_attr = custom_logger_factory.rsplit(".", 1)
                if module_name in sys.modules:
                    logger_factory = sys.modules[module_name].__dict__.get(module_attr)
                else:
                    logger_factory = __import__(
                        module_name, globals=None, locals=None, fromlist=(module_attr), level=0
                    ).__dict__.get(module_attr)
            else:
                module_name = custom_logger_factory
                if module_name in sys.modules:
                    logger_factory = sys.modules[module_name]
                else:
                    logger_factory = __import__(module_name, globals=None, locals=None, fromlist=(), level=0)

            if not logger_factory:
                raise Exception("Invalid path to custom logger")
        else:
            logger_factory = custom_logger_factory

        if not any(
            [
                (isinstance(logger_factory, ModuleType) and logger_factory.__name__ == "logging"),
                isinstance(logger_factory, type),
                hasattr(logger_factory, "new"),
                hasattr(logger_factory, "bind"),
                hasattr(logger_factory, "get_logger"),
            ]
        ):
            raise Exception("Invalid custom logger factory")

        if isinstance(custom_logger_factory, str):
            cls._logger_factory_cache[custom_logger_factory] = logger_factory

        return logger_factory

    @classmethod
    def get_logger(cls, context: Union[LoggerContext, Context]) -> Any:
        if not TOMODACHI_CUSTOM_LOGGER:
            raise Exception("Custom logger not set")

        logger_factory: Any

        if isinstance(TOMODACHI_CUSTOM_LOGGER, str):
            logger_factory = cls.get_logger_factory(TOMODACHI_CUSTOM_LOGGER or "")
        else:
            logger_factory = TOMODACHI_CUSTOM_LOGGER

        if isinstance(logger_factory, ModuleType) and logger_factory.__name__ == "logging":
            return logger_factory.getLogger(context.get("logger", "default"))
        elif isinstance(logger_factory, type):
            return logger_factory(context)
        elif hasattr(logger_factory, "new"):
            return logger_factory.new(**context)
        elif hasattr(logger_factory, "bind"):
            return logger_factory.bind(**context)
        elif hasattr(logger_factory, "get_logger"):
            logger_cls = logger_factory.get_logger()
            if hasattr(logger_cls, "new"):
                return logger_cls.new(**context)

        raise Exception("Invalid custom logger factory")


class Logger(structlog.stdlib.BoundLogger):
    def __init__(
        self,
        logger: WrappedLogger,
        processors: Iterable[Processor],
        context: Union[LoggerContext, Context],
    ):
        if (
            logger
            and isinstance(logger, logging.Logger)
            and logger.name == "default"
            and context.get("logger", "default") != "default"
        ):
            logger = logging.getLogger(context.get("logger"))

        if logger is _CustomLogger:
            logger = _CustomLogger.get_logger(context)

        super().__init__(logger, processors, context)

    def _proxy_to_logger(self, method_name: str, event: Optional[str] = None, *event_args: Any, **event_kw: Any) -> Any:
        if method_name == "error" and event_kw and event_kw.get("exc_info") is True:
            method_name = "exception"

        try:
            if not self.isEnabledFor(_NAME_TO_LEVEL[method_name]):
                return None
        except Exception:
            try:
                level = self.level
            except Exception:
                try:
                    level = logging.root.level
                except Exception:
                    level = NOTSET

            try:
                if level > _NAME_TO_LEVEL[method_name]:
                    return None
            except Exception:
                pass

        return super()._proxy_to_logger(method_name, event, *event_args, **event_kw)

    @property
    def name(self) -> str:
        name: str = self._context.get("logger") or ""
        if not name:
            return str(self._logger.name or "")
        return name

    @property
    def level(self) -> int:
        try:
            return int(self._logger.level)
        except AttributeError:
            name = self._context.get("logger") or None
            return logging.getLogger(name).level or logging.root.level

    @property
    def parent(self) -> Any:
        try:
            return self._logger.parent
        except AttributeError:
            name = self._context.get("logger") or None
            return logging.getLogger(name).parent

    @property
    def propagate(self) -> bool:
        try:
            return bool(self._logger.propagate)
        except AttributeError:
            name = self._context.get("logger") or None
            return logging.getLogger(name).propagate

    @property
    def handlers(self) -> Any:
        try:
            return self._logger.handlers
        except AttributeError:
            name = self._context.get("logger") or None
            return logging.getLogger(name).handlers

    @property
    def disabled(self) -> int:
        try:
            return int(bool(self._logger.disabled))
        except AttributeError:
            name = self._context.get("logger") or None
            return logging.getLogger(name).disabled

    def setLevel(self, level: int) -> None:
        name = self._context.get("logger") or None
        logging.getLogger(name).setLevel(level)

        try:
            self._logger.setLevel(level)
        except AttributeError:
            pass

    def findCaller(self, stack_info: bool = False) -> Tuple[str, int, str, Optional[str]]:
        try:
            return cast(Tuple[str, int, str, Optional[str]], self._logger.findCaller(stack_info=stack_info))  # type: ignore
        except AttributeError:
            name = self._context.get("logger") or None
            return logging.getLogger(name).findCaller(stack_info=stack_info)

    def makeRecord(
        self,
        name: str,
        level: int,
        fn: str,
        lno: int,
        msg: str,
        args: tuple[Any, ...],
        exc_info: ExcInfo,
        func: str | None = None,
        extra: Any = None,
    ) -> logging.LogRecord:
        try:
            return cast(  # type: ignore
                logging.LogRecord,
                self._logger.makeRecord(name, level, fn, lno, msg, args, exc_info, func=func, extra=extra),
            )
        except AttributeError:
            name_ = self._context.get("logger") or None
            return logging.getLogger(name_).makeRecord(
                name, level, fn, lno, msg, args, exc_info, func=func, extra=extra
            )

    def handle(self, record: logging.LogRecord) -> None:
        try:
            self._logger.handle(record)
        except AttributeError:
            name = self._context.get("logger") or None
            logging.getLogger(name).handle(record)

    def addHandler(self, hdlr: logging.Handler) -> None:
        try:
            self._logger.addHandler(hdlr)
        except AttributeError:
            name = self._context.get("logger") or None
            logging.getLogger(name).addHandler(hdlr)

    def removeHandler(self, hdlr: logging.Handler) -> None:
        try:
            self._logger.removeHandler(hdlr)
        except AttributeError:
            name = self._context.get("logger") or None
            logging.getLogger(name).removeHandler(hdlr)

    def hasHandlers(self) -> bool:
        try:
            return bool(self._logger.hasHandlers())
        except AttributeError:
            name = self._context.get("logger") or None
            return logging.getLogger(name).hasHandlers()

    def callHandlers(self, record: logging.LogRecord) -> None:
        try:
            self._logger.callHandlers(record)
        except AttributeError:
            name = self._context.get("logger") or None
            logging.getLogger(name).callHandlers(record)

    def getEffectiveLevel(self) -> int:
        try:
            return int(self._logger.getEffectiveLevel())
        except AttributeError:
            name = self._context.get("logger") or None
            return logging.getLogger(name).getEffectiveLevel()

    def isEnabledFor(self, level: int) -> bool:
        try:
            return bool(self._logger.isEnabledFor(level))
        except AttributeError:
            name = self._context.get("logger") or None
            return logging.getLogger(name).isEnabledFor(level)

    def getChild(self, suffix: str) -> logging.Logger:
        try:
            return cast(logging.Logger, self._logger.getChild(suffix))  # type: ignore
        except AttributeError:
            name = self._context.get("logger") or None
            return logging.getLogger(name).getChild(suffix)


class LoggerProtocol(Protocol):
    def info(self, *args: Any, **kwargs: Any) -> None: ...

    def debug(self, *args: Any, **kwargs: Any) -> None: ...

    def warning(self, *args: Any, **kwargs: Any) -> None: ...

    def error(self, *args: Any, **kwargs: Any) -> None: ...

    def critical(self, *args: Any, **kwargs: Any) -> None: ...

    def exception(self, *args: Any, **kwargs: Any) -> None: ...


ConsoleRenderer: type[structlog.dev.ConsoleRenderer]
_structlog_version = tuple(map(lambda v: int(v) if v.isdigit() else v, structlog.__version__.split(".")))
if _structlog_version < (23, 3, 0):  # pragma: no cover
    # backport of structlog 23.x ConsoleRenderer to be usable with structlog 21.x+.

    class ConsoleRenderer_(structlog.dev.ConsoleRenderer):
        _colors: bool
        _level_to_color: Dict[str, str]
        _pad_event: int
        _event_key: str

        def __init__(
            self,
            **kw: Any,
        ):
            try:
                super().__init__(**kw)
                self._event_key = kw.get("event_key") or "event"
            except Exception:
                self._event_key = kw.pop("event_key", "event")
                super().__init__(**kw)

            self._colors = kw.get("colors", not NO_COLOR) and not NO_COLOR
            self._pad_event = kw.get("pad_event") or structlog.dev._EVENT_WIDTH

            level_styles = kw.get("level_styles")
            level_to_color = self.get_default_level_styles(self._colors) if level_styles is None else level_styles
            for key in level_to_color:
                level_to_color[key] += self._styles.bright
            self._level_to_color = level_to_color

        def __call__(self, logger: WrappedLogger, method_name: str, event_dict: EventDict) -> str:
            sio = StringIO()

            _styles = self._styles
            _level_to_color = self._level_to_color

            ts = event_dict.pop("timestamp", None)
            if ts is not None:
                sio.write(_styles.timestamp + str(ts) + _styles.reset + " ")
            level = event_dict.pop("level", None)
            if level is not None:
                sio.write(
                    "["
                    + _level_to_color.get(level, "")
                    + structlog.dev._pad(level, self._longest_level)
                    + _styles.reset
                    + "] "
                )

            # force event to str for compatibility with standard library
            event = event_dict.pop(self._event_key, None)
            if not isinstance(event, str):
                event = str(event)

            if event_dict:
                event = structlog.dev._pad(event, self._pad_event) + _styles.reset + " "
            else:
                event += _styles.reset
            sio.write(_styles.bright + event)

            logger_name = event_dict.pop("logger", None)
            if logger_name is None:
                logger_name = event_dict.pop("logger_name", None)

            if logger_name is not None:
                sio.write("[" + _styles.logger_name + _styles.bright + logger_name + _styles.reset + "] ")

            stack = event_dict.pop("stack", None)
            exc = event_dict.pop("exception", None)
            exc_info = event_dict.pop("exc_info", None)

            event_dict_keys: Iterable[str] = event_dict.keys()
            if self._sort_keys:
                event_dict_keys = sorted(event_dict_keys)

            sio.write(
                " ".join(
                    _styles.kv_key
                    + key
                    + _styles.reset
                    + "="
                    + _styles.kv_value
                    + self._repr(event_dict[key])
                    + _styles.reset
                    for key in event_dict_keys
                )
            )

            if stack is not None:
                sio.write("\n" + stack)
                if exc_info or exc is not None:
                    sio.write("\n\n" + "=" * 79 + "\n")

            if exc_info:
                exc_info = structlog.processors._figure_out_exc_info(exc_info)

                self._exception_formatter(sio, exc_info)
            elif exc is not None:
                if self._exception_formatter is not structlog.dev.plain_traceback:
                    warnings.warn(
                        "Remove `format_exc_info` from your processor chain " "if you want pretty exceptions.",
                        stacklevel=2,
                    )
                sio.write("\n" + exc)

            return sio.getvalue()

    ConsoleRenderer = ConsoleRenderer_
else:
    ConsoleRenderer = structlog.dev.ConsoleRenderer

console_logger: Logger = structlog.wrap_logger(
    None,
    processors=[
        structlog.processors.add_log_level,
        merge_contextvars,
        structlog.processors.StackInfoRenderer(),
        structlog.dev.set_exc_info,
        LogProcessorTimestamp(),
        RenameKeys(pairs=RENAME_KEYS),
        add_exception_info,
        AddMissingDictKey(key="message"),
        remove_ellipsis_values,
        LinkQuoteStrings(keys=CONSOLE_QUOTE_KEYS),
        ConsoleRenderer(colors=False if NO_COLOR else True, sort_keys=False, event_key="message"),
    ],
    wrapper_class=Logger,
    context_class=LoggerContext,
    cache_logger_on_first_use=False,
)

no_color_console_logger: Logger = structlog.wrap_logger(
    None,
    processors=[
        structlog.processors.add_log_level,
        merge_contextvars,
        structlog.processors.StackInfoRenderer(),
        structlog.dev.set_exc_info,
        LogProcessorTimestamp(),
        RenameKeys(pairs=RENAME_KEYS),
        add_exception_info,
        AddMissingDictKey(key="message"),
        remove_ellipsis_values,
        LinkQuoteStrings(keys=CONSOLE_QUOTE_KEYS),
        ConsoleRenderer(colors=False, sort_keys=False, event_key="message"),
    ],
    wrapper_class=Logger,
    context_class=LoggerContext,
    cache_logger_on_first_use=False,
)

json_logger: Logger = structlog.wrap_logger(
    None,
    processors=[
        structlog.processors.add_log_level,
        merge_contextvars,
        structlog.processors.StackInfoRenderer(),
        structlog.dev.set_exc_info,
        LogProcessorTimestamp(),
        RenameKeys(pairs=RENAME_KEYS),
        add_exception_info,
        add_stacktrace_info,
        RemoveDictKey(key="exc_info"),
        remove_ellipsis_values,
        structlog.processors.JSONRenderer(serializer=serializer_func),
    ],
    wrapper_class=Logger,
    context_class=LoggerContext,
    cache_logger_on_first_use=False,
)

custom_logger: Logger = structlog.wrap_logger(
    _CustomLogger,
    processors=[remove_ellipsis_values, to_custom_logger],
    wrapper_class=Logger,
    context_class=LoggerContext,
    cache_logger_on_first_use=False,
)

forward_logger: Logger = structlog.wrap_logger(
    logging.getLogger("default"),
    processors=[remove_ellipsis_values, to_logger_args_kwargs],
    wrapper_class=Logger,
    context_class=LoggerContext,
    cache_logger_on_first_use=False,
)

python_logger: Logger = structlog.wrap_logger(
    logging.getLogger("default"),
    processors=[remove_ellipsis_values, to_logger_args_kwargs],
    wrapper_class=Logger,
    context_class=LoggerContext,
    cache_logger_on_first_use=False,
)

disabled_logger: Logger = structlog.wrap_logger(
    None,
    processors=[disabled_logger_processor],
    wrapper_class=Logger,
    context_class=LoggerContext,
    cache_logger_on_first_use=False,
)


def get_context(logger: Union[LoggerContext, Dict, structlog.BoundLoggerBase, str]) -> Union[LoggerContext, Dict]:
    if isinstance(logger, structlog.BoundLoggerBase):
        if not isinstance(logger._context, LoggerContext):
            return cast(dict, logger._context)
        return logger._context
    elif isinstance(logger, (LoggerContext, dict)):
        return logger
    elif isinstance(logger, str):
        ctx = _loggers.get().get(logger)
        if not ctx:
            ctx = LoggerContext(logger=logger, **{})
        return ctx
    else:
        raise TypeError(f"Unsupported logger type: {type(logger)}")


def bind_logger(logger: Union[LoggerContext, Dict, structlog.BoundLoggerBase, str]) -> None:
    _context.set(get_context(logger))


def _get_logger(
    name: Optional[str] = None,
    *,
    logger_type: Literal["json", "console", "no_color_console", "custom", "forward", "python", "disabled"] = "forward",
) -> Logger:
    if logger_type == "forward" and TOMODACHI_LOGGER_TYPE == "python":
        logger_type = "python"

    logger = (
        json_logger
        if logger_type == "json"
        else (
            console_logger
            if logger_type == "console"
            else (
                no_color_console_logger
                if logger_type == "no_color_console"
                else (
                    custom_logger
                    if logger_type == "custom"
                    else (
                        python_logger
                        if logger_type == "python"
                        else (
                            disabled_logger
                            if logger_type == "disabled"
                            else forward_logger if logger_type == "forward" else None
                        )
                    )
                )
            )
        )
    )
    if not logger:
        raise Exception(
            "Invalid logger type: '{}' (exected 'console', 'json', 'custom', 'forward', 'python' or 'disabled')".format(
                logger_type
            )
        )

    if name:
        ctx = _loggers.get().get(name)
        if ctx:
            return cast(Logger, logger.new(**ctx))

        return cast(Logger, logger.new(**LoggerContext(logger=name, **{})))

    return cast(Logger, logger.new(**_context.get()))


def get_logger(
    name: Optional[str] = None,
) -> Logger:
    return _get_logger(name=name)


# CamelCase alias for `get_logger`.
def getLogger(name: Optional[str] = None) -> Logger:
    return _get_logger(name=name)


def configure(log_level: Union[int, str] = logging.INFO, force: bool = False) -> None:
    log_level_ = log_level
    if isinstance(log_level, str) and not log_level.isdigit():
        log_level_ = getattr(logging, log_level.upper(), None) or logging.NOTSET
    if isinstance(log_level, str) and log_level.isdigit():
        log_level_ = int(log_level)
    if type(log_level_) is not int or log_level_ == logging.NOTSET:
        raise Exception(
            f"Invalid log level: '{log_level}' (expected 'debug', 'info', 'warning', 'error' or 'critical')"
        )
    log_level = log_level_

    logging.addLevelName(logging.NOTSET, "notset")
    logging.addLevelName(logging.DEBUG, "debug")
    logging.addLevelName(logging.INFO, "info")
    logging.addLevelName(logging.WARN, "warn")
    logging.addLevelName(logging.WARNING, "warning")
    logging.addLevelName(logging.ERROR, "error")
    logging.addLevelName(logging.FATAL, "fatal")
    logging.addLevelName(logging.CRITICAL, "critical")

    global _default_fmt
    try:
        if (
            len(logging.root.handlers)
            and logging.root.handlers[0].formatter
            and logging.root.handlers[0].formatter._fmt
        ):
            _default_fmt = logging.root.handlers[0].formatter._fmt
    except AttributeError:
        pass

    if PythonLoggingFormatter._fmt != _default_fmt:
        PythonLoggingFormatter._fmt = _default_fmt

    try:
        logging.basicConfig(
            format=_default_fmt,
            level=log_level,
            handlers=[DefaultRootLoggerHandler],
            force=force,
        )
    except Exception as e:
        logging.getLogger().warning("Unable to set log config: {}".format(str(e)))


def remove_handlers() -> None:
    for name, logger in logging.Logger.manager.loggerDict.items():
        if name not in ("tomodachi",) and not name.startswith("tomodachi."):
            continue

        if isinstance(logger, logging.PlaceHolder):
            continue

        if not getattr(logger, "handlers", None) or not logger.handlers or not isinstance(logger.handlers, list):
            continue

        for handler in logger.handlers:
            logger.removeHandler(handler)
            try:
                handler.acquire()
                handler.flush()
                handler.close()
            except (OSError, ValueError):
                pass
            finally:
                handler.release()

    for handler in logging.root.handlers:
        if handler is not DefaultRootLoggerHandler:
            continue
        logging.root.removeHandler(handler)
        try:
            handler.acquire()
            handler.flush()
            handler.close()
        except (OSError, ValueError):
            pass
        finally:
            handler.release()


# Set default logger context
_context.set(LoggerContext(logger="default", **{}))


def reset_context() -> None:
    _loggers.set({})
    _context.set(LoggerContext(logger="default", **{}))


__all__ = [
    "get_logger",
    "_get_logger",
    "getLogger",
    "bind_logger",
    "reset_context",
    "Logger",
    "PythonLoggingFormatter",
    "ConsoleFormatter",
    "NoColorConsoleFormatter",
    "JSONFormatter",
    "DisabledFormatter",
    "CustomLoggerFormatter",
    "DefaultHandler",
    "DefaultRootLoggerHandler",
    "_defaultHandler",
    "StderrHandler",
    "configure",
    "set_default_formatter",
    "remove_handlers",
    "add_exception_info",
    "add_stacktrace_info",
    "DEFAULT_FORMAT",
    "CRITICAL",
    "DEBUG",
    "ERROR",
    "FATAL",
    "INFO",
    "NOTSET",
    "WARN",
    "WARNING",
]



import os

NO_COLOR = any(
    [
        os.environ.get("NO_COLOR", "").lower() in ("1", "true"),
        os.environ.get("NOCOLOR", "").lower() in ("1", "true"),
        os.environ.get("TOMODACHI_NO_COLOR", "").lower() in ("1", "true"),
        os.environ.get("TOMODACHI_NOCOLOR", "").lower() in ("1", "true"),
        os.environ.get("CLICOLOR", "").lower() in ("0", "false"),
        os.environ.get("CLI_COLOR", "").lower() in ("0", "false"),
        os.environ.get("CLICOLOR_FORCE", "").lower() in ("0", "false"),
    ]
)


class ColorFore:
    BLACK = ""
    RED = ""
    GREEN = ""
    YELLOW = ""
    BLUE = ""
    MAGENTA = ""
    CYAN = ""
    WHITE = ""
    RESET = ""

    LIGHTBLACK_EX = ""
    LIGHTRED_EX = ""
    LIGHTGREEN_EX = ""
    LIGHTYELLOW_EX = ""
    LIGHTBLUE_EX = ""
    LIGHTMAGENTA_EX = ""
    LIGHTCYAN_EX = ""
    LIGHTWHITE_EX = ""


class ColorStyle:
    BRIGHT = ""
    DIM = ""
    NORMAL = ""
    RESET_ALL = ""


COLOR_0 = COLOR = ColorFore()
COLOR_STYLE_0 = COLOR_STYLE = ColorStyle()
COLOR_RESET_0 = COLOR_RESET = ""
try:
    import colorama  # noqa  # isort:skip

    if not NO_COLOR:
        COLOR = colorama.Fore
        COLOR_STYLE = colorama.Style
        COLOR_RESET = colorama.Style.RESET_ALL
except Exception:
    pass

__all__ = ["COLOR", "COLOR_STYLE", "COLOR_RESET", "COLOR_0", "COLOR_STYLE_0", "COLOR_RESET_0", "NO_COLOR"]



import asyncio
import inspect
import time
from contextlib import AsyncExitStack, asynccontextmanager
from typing import Any, AsyncIterator, Dict, Optional, cast

import aiobotocore
import aiobotocore.client
import aiobotocore.config
import aiobotocore.session
import aiohttp
import aiohttp.client_exceptions
import botocore
import botocore.exceptions

MAX_POOL_CONNECTIONS = 50
CONNECT_TIMEOUT = 8
READ_TIMEOUT = 35
CLIENT_CREATION_TIME_LOCK = 45


class ClientConnector(object):
    __slots__ = (
        "clients",
        "_clients_context",
        "credentials",
        "client_creation_lock_time",
        "aliases",
        "locks",
        "conditions",
        "close_waiter",
    )

    clients: Dict[str, Optional[aiobotocore.client.AioBaseClient]]
    _clients_context: Dict[str, AsyncExitStack]
    credentials: Dict[str, Dict]
    client_creation_lock_time: Dict[str, float]
    aliases: Dict[str, str]
    locks: Dict[str, asyncio.Lock]
    conditions: Dict[str, asyncio.Condition]
    close_waiter: Optional[asyncio.Future]

    def __init__(self) -> None:
        self.clients = {}
        self._clients_context = {}
        self.credentials = {}
        self.aliases = {}
        self.client_creation_lock_time = {}
        self.locks = {}
        self.conditions = {}
        self.close_waiter = None

    def setup_credentials(self, alias_name: str, credentials: Dict) -> None:
        self.credentials[alias_name] = credentials

    def get_client(self, alias_name: str) -> Optional[aiobotocore.client.AioBaseClient]:
        return self.clients.get(alias_name)

    def get_lock(self, alias_name: str) -> asyncio.Lock:
        if alias_name not in self.locks:
            self.locks[alias_name] = asyncio.Lock()
        return self.locks[alias_name]

    async def get_condition(self, alias_name: str) -> asyncio.Condition:
        if alias_name not in self.conditions:
            self.conditions[alias_name] = asyncio.Condition()
        return self.conditions[alias_name]

    async def create_client(
        self, alias_name: Optional[str] = None, credentials: Optional[Dict] = None, service_name: Optional[str] = None
    ) -> aiobotocore.client.AioBaseClient:
        if service_name is None and alias_name is not None:
            service_name = alias_name
        if alias_name is None and service_name is not None:
            alias_name = service_name

        alias_name = str(alias_name)
        service_name = str(service_name)

        exc_iteration_count = 0
        while self.close_waiter:
            if self.close_waiter.done():
                self.close_waiter = None
            else:
                try:
                    await self.close_waiter
                except RuntimeError:
                    exc_iteration_count += 1
                    if exc_iteration_count >= 100:
                        raise
                    await asyncio.sleep(0.1)

        async with self.get_lock(alias_name):
            client = self.get_client(alias_name)
            if self.client_creation_lock_time.get(alias_name, 0) + CLIENT_CREATION_TIME_LOCK > time.time():
                if client:
                    return client
            self.client_creation_lock_time[alias_name] = time.time() if client else 0

            if not credentials:
                credentials = self.credentials.get(alias_name, {})
            else:
                self.credentials[alias_name] = credentials
            if not credentials:
                credentials = {}

            self.aliases[alias_name] = service_name

            session = aiobotocore.session.get_session()
            config = aiobotocore.config.AioConfig(
                connect_timeout=CONNECT_TIMEOUT, read_timeout=READ_TIMEOUT, max_pool_connections=MAX_POOL_CONNECTIONS
            )
            context_stack = AsyncExitStack()
            client_value = context_stack.enter_async_context(
                session.create_client(service_name, config=config, **credentials)
            )
            if inspect.isawaitable(client_value):
                client = await client_value
            else:
                client = client_value

            old_client = self.get_client(alias_name)
            self.clients[alias_name] = cast(aiobotocore.client.AioBaseClient, client)
            self._clients_context[alias_name] = context_stack

            if old_client:
                try:
                    await asyncio.sleep(1)
                    task = old_client.close()
                    if getattr(task, "_coro", None):
                        task = getattr(task, "_coro")
                    await asyncio.wait([asyncio.ensure_future(task)], timeout=3)
                    await asyncio.sleep(0.25)  # SSL termination sleep
                except (Exception, RuntimeError, asyncio.CancelledError, BaseException):
                    pass

            return cast(aiobotocore.client.AioBaseClient, self.clients.get(alias_name) or client)

    async def close_client(
        self,
        alias_name: Optional[str] = None,
        client: Optional[aiobotocore.client.AioBaseClient] = None,
        fast: bool = False,
    ) -> None:
        exc_iteration_count = 0
        while self.close_waiter:
            if self.close_waiter.done():
                self.close_waiter = None
            else:
                try:
                    await self.close_waiter
                except RuntimeError:
                    exc_iteration_count += 1
                    if exc_iteration_count >= 100:
                        raise
                    await asyncio.sleep(0.1)

        if not alias_name and client:
            if client in self.clients.values():
                alias_name = [k for k, v in self.clients.items() if client == v][0]

        if not alias_name:
            return

        async with self.get_lock(alias_name):
            client = self.get_client(alias_name)
            context_stack = self._clients_context.get(alias_name)
            if not client or not context_stack:
                return

            try:
                if not fast:
                    await asyncio.sleep(0.25)
                task = client.close()
                if getattr(task, "_coro", None):
                    task = getattr(task, "_coro")
                await asyncio.wait(
                    [asyncio.ensure_future(task), asyncio.ensure_future(context_stack.aclose())], timeout=3
                )
                if not fast:
                    await asyncio.sleep(0.25)  # SSL termination sleep
                else:
                    await asyncio.sleep(0)
            except (Exception, RuntimeError, asyncio.CancelledError, BaseException):
                pass

            self.clients[alias_name] = None

    async def reconnect_client(
        self, alias_name: Optional[str] = None, client: Optional[aiobotocore.client.AioBaseClient] = None
    ) -> aiobotocore.client.AioBaseClient:
        if not alias_name and client:
            if client in self.clients.values():
                alias_name = [k for k, v in self.clients.items() if client == v][0]

        service_name = self.aliases.get(alias_name or "")
        if not service_name:
            raise Exception("Client has never been created in the first place")

        return await self.create_client(alias_name, service_name=service_name)

    async def close(self, fast: bool = False) -> None:
        if self.close_waiter and not self.close_waiter.done():
            return

        clients = self.clients

        self.clients = {}
        self.credentials = {}
        self.aliases = {}
        self.client_creation_lock_time = {}
        self.locks = {}

        if not clients:
            return

        self.close_waiter = asyncio.Future()
        if not fast:
            await asyncio.sleep(0.25)

        tasks = []
        for _, client in clients.items():
            if not client:
                continue
            try:
                task = client.close()
                if getattr(task, "_coro", None):
                    task = getattr(task, "_coro")
                tasks.append(asyncio.ensure_future(task))
            except (Exception, RuntimeError, asyncio.CancelledError, BaseException):
                pass

        try:
            await asyncio.wait(tasks, timeout=3)
            if not fast:
                await asyncio.sleep(0.25)  # SSL termination sleep
            else:
                await asyncio.sleep(0)
        except (Exception, RuntimeError, asyncio.CancelledError, BaseException):
            pass

        if self.close_waiter:
            self.close_waiter.set_result(None)

    @asynccontextmanager
    async def __call__(
        self, alias_name: Optional[str] = None, credentials: Optional[Dict] = None, service_name: Optional[str] = None
    ) -> AsyncIterator[Any]:
        exc_iteration_count = 0
        while self.close_waiter and not self.close_waiter.done():
            try:
                await self.close_waiter
            except RuntimeError:
                exc_iteration_count += 1
                if exc_iteration_count >= 100:
                    raise
                await asyncio.sleep(0.1)

        client_name = alias_name or service_name or ""

        if not self.get_client(client_name):
            await self.create_client(alias_name, credentials, service_name)
        client = self.get_client(client_name)

        try:
            yield client
        except (
            botocore.exceptions.NoRegionError,
            botocore.exceptions.PartialCredentialsError,
            botocore.exceptions.NoCredentialsError,
        ):
            await self.close_client(client=client, fast=True)
            raise
        except (aiohttp.client_exceptions.ServerDisconnectedError, asyncio.TimeoutError, RuntimeError):
            await self.reconnect_client(client_name, client=client)
            raise
        except botocore.exceptions.ClientError as e:
            error_message = str(e)
            if "The security token included in the request is invalid" in error_message:
                await self.close_client(client=client, fast=True)
            raise


connector: ClientConnector = ClientConnector()
